from __future__ import annotations
from tuneflow_py.models.protos import song_pb2


class Note:
    def __init__(self, pitch: int | None = None, velocity: int | None = None, start_tick: int | None = None, end_tick: int |
                 None = None, id: int | None = None, clip=None, proto: song_pb2.Note | None = None) -> None:
        '''
        IMPORTANT: Do not use the constructor directly, call createNote from clips instead.

        @param clip  If left empty, all clip-related methods will be no-op.
        '''
        self.clip = clip
        if proto is not None:
            self._proto = proto
        else:
            self._proto = song_pb2.Note()
            self._proto.pitch = pitch
            self._proto.velocity = velocity
            self._proto.start_tick = start_tick
            self._proto.end_tick = end_tick
            self._proto.id = id

    def get_id(self) -> int:
        return self._proto.id

    def get_pitch(self) -> int:
        return self._proto.pitch

    def get_velocity(self) -> int:
        return self._proto.velocity

    def set_velocity(self, velocity: int):
        self._proto.velocity = velocity

    def get_start_tick(self) -> int:
        return self._proto.start_tick

    def set_start_tick(self, start_tick: int):
        self._proto.start_tick = start_tick

    def get_end_tick(self) -> int:
        return self._proto.end_tick

    def set_end_tick(self, end_tick: int):
        self._proto.end_tick = end_tick

    def set_pitch(self, pitch: int):
        if not Note.is_valid_pitch(pitch):
            raise Exception("Invalid note pitch " + str(pitch))
        self._proto.pitch = pitch

    def adjust_pitch(self, pitch_offset: int):
        '''
        Adjust the pitch of a note.

        If the pitch of the note becomes invalid (less than 0 or greater than 127),
        it will be deleted from the clip.
        '''
        self._proto.pitch = self._proto.pitch + pitch_offset
        if not Note.is_valid_pitch(self._proto.pitch):
            self.delete_from_parent()

    def get_start_time(self) -> float:
        return self._proto.start_time

    def get_end_time(self) -> float:
        return self._proto.end_time

    def get_clip(self):
        return self.clip

    def adjust_left(self, offset_tick: int):
        '''
        Adjusts the start tick of the note by an offset.
        '''
        if (offset_tick == 0):
            return

        clip = self.clip
        if clip is not None:
            self.delete_from_parent()

        self._proto.start_tick += offset_tick
        if (not self.is_range_valid()):
            # Note is out of valid range, delete it
            # by not inserting it back.
            return

        if (clip is not None):
            clip._ordered_insert_note(self)

    def adjust_left_to(self, tick: int):
        '''
        Adjusts the start tick of the note to a given tick.
        '''
        self.adjust_left(tick - self.get_start_tick())

    def adjust_right(self, offset_tick: int):
        '''
        Adjusts the end tick of the note by an offset.
        '''
        self._proto.end_tick += offset_tick
        if (not self.is_range_valid()):
            self.delete_from_parent()

    def adjust_right_to(self, tick: int):
        '''
        Adjusts the end tick of the note to a given tick.
        '''
        self.adjust_right(tick - self.get_end_tick())

    def is_range_valid(self):
        return Note.is_note_range_valid(self._proto.start_tick, self._proto.end_tick)

    def move_note(self, offset_tick: int):
        '''
        Moves the note by a given offset in terms of ticks.

        This will not update the clip's boundaries.
        '''
        if (offset_tick == 0):
            return
        clip = self.clip
        if clip is not None:
            self.delete_from_parent()

        self._proto.start_tick = max(0, self._proto.start_tick + offset_tick)
        self._proto.end_tick = self._proto.end_tick + offset_tick
        if not self.is_range_valid():
            # Note is out of valid range, delete it
            # by not inserting it back.
            return
        if clip is not None:
            clip._ordered_insert_note(self)

    def delete_from_parent(self):
        if self.clip is None:
            return

        self.clip.delete_note(self)
        self.clip = None

    def equals(self, note: Note):
        '''
        Returns true if the notes should sound the same.

        NOTE: This does not check note Ids or the clips they belong to.
        '''
        return (
            self.get_start_tick() == note.get_start_tick() and
            self.get_end_tick() == note.get_end_tick() and
            self.get_pitch() == note.get_pitch() and
            self.get_velocity() == note.get_velocity()
        )

    def __repr__(self) -> str:
        return str(self._proto)

    @staticmethod
    def is_valid_pitch(pitch: int):
        return pitch >= 0 and pitch <= 127 and isinstance(pitch, int)

    @staticmethod
    def is_note_range_valid(start_tick: int, end_tick: int):
        return (
            end_tick >= 0 and
            start_tick <= end_tick
        )

    @staticmethod
    def is_note_velocity_valid(velocity: int):
        return velocity >= 0 and velocity <= 127



from tuneflow_py import Song, Clip, TrackType, Note
from typing import List
import unittest


def create_song():
    song = Song()
    song.create_tempo_change(ticks=1440, bpm=60)
    track = song.create_track(type=TrackType.MIDI_TRACK)
    clip1 = track.create_midi_clip(clip_start_tick=0)
    clip1.create_note(pitch=64, velocity=80, start_tick=0, end_tick=10)
    clip1.create_note(pitch=66, velocity=80, start_tick=15, end_tick=20)
    clip1.create_note(pitch=68, velocity=80, start_tick=14, end_tick=20)
    clip1.adjust_clip_left(0)
    clip1.adjust_clip_right(15)
    clip2 = track.create_midi_clip(clip_start_tick=21)
    clip2.create_note(pitch=66, velocity=80, start_tick=25, end_tick=30)
    clip2.create_note(pitch=64, velocity=80, start_tick=18, end_tick=35)
    clip2.create_note(pitch=68, velocity=80, start_tick=24, end_tick=35)
    clip2.adjust_clip_left(21)
    clip2.adjust_clip_right(30)
    clip3 = track.create_midi_clip(clip_start_tick=40)
    clip3.create_note(pitch=71, velocity=80, start_tick=55, end_tick=65)
    clip3.create_note(pitch=69, velocity=80, start_tick=45, end_tick=50)
    clip3.create_note(pitch=67, velocity=80, start_tick=40, end_tick=45)
    clip3.adjust_clip_left(40)
    clip3.adjust_clip_right(65)
    return song


def assert_notes_are_equal(notes1: List[Note], notes2: List[Note]):
    if len(notes1) != len(notes2):
        raise Exception(
            f'Notes are not of equal length, {len(notes1)} vs {len(notes2)}')
    for i in range(0, len(notes1)):
        if not notes1[i].equals(notes2[i]) or notes1[i].get_id() != notes2[i].get_id():
            raise Exception(
                f'{i}th notes of the two lists are not equal, \n{str(notes1[i])}\n vs \n{str( notes2[i],)}',
            )


def create_test_notes(note_specs: List, clip: Clip):
    return [Note(**spec, clip=clip) for spec in note_specs]


class BaseTestCase(unittest.TestCase):
    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        song = create_song()
        self.song = song

    def assert_clip_range(self, clip: Clip, start_tick: int, end_tick: int):
        self.assertEqual(clip.get_clip_start_tick(), start_tick)
        self.assertEqual(clip.get_clip_end_tick(), end_tick)

# class TestIsNoteInClip(BaseTestCase):
#     def test_is_note_in_clip(self):
#         self.assertFalse(Clip.is_note_in_clip(note_start_tick=4, note_end_tick=5,clip_start_tick=6,clip_end_tick=12))
#         self.assertFalse(Clip.is_note_in_clip(note_start_tick=5, note_end_tick=10,clip_start_tick=6,clip_end_tick=12))
#         self.assertTrue(Clip.is_note_in_clip(note_start_tick=6, note_end_tick=10,clip_start_tick=6,clip_end_tick=12))
#         self.assertTrue(Clip.is_note_in_clip(note_start_tick=6, note_end_tick=12,clip_start_tick=6,clip_end_tick=12))
#         self.assertTrue(Clip.is_note_in_clip(note_start_tick=7, note_end_tick=12,clip_start_tick=6,clip_end_tick=12))
#         self.assertTrue(Clip.is_note_in_clip(note_start_tick=7, note_end_tick=13,clip_start_tick=6,clip_end_tick=12))
#         self.assertFalse(Clip.is_note_in_clip(note_start_tick=12, note_end_tick=13,clip_start_tick=6,clip_end_tick=12))
#         self.assertFalse(Clip.is_note_in_clip(note_start_tick=13, note_end_tick=15,clip_start_tick=6,clip_end_tick=12))
#         self.assertFalse(Clip.is_note_in_clip(note_start_tick=5, note_end_tick=15,clip_start_tick=6,clip_end_tick=12))
#         self.assertTrue(Clip.is_note_in_clip(note_start_tick=-13, note_end_tick=5,clip_start_tick=0,clip_end_tick=12))


class TestBasicOperations(BaseTestCase):
    def test_gets_clip_ranges(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assertEqual(clip1.get_clip_start_tick(), 0)
        self.assertEqual(clip1.get_clip_end_tick(), 15)

    def test_get_notes_within_clip_range(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        clip2 = track.get_clip_at(1)
        clip3 = track.get_clip_at(2)
        # Notes that are at the edge are included.
        assert_notes_are_equal(
            list(clip1.get_notes()),
            create_test_notes(
                [
                    {
                        "id": 1,
                        "pitch": 64,
                        "velocity": 80,
                        "start_tick": 0,
                        "end_tick": 10,
                    },
                    {
                        "id": 3,
                        "pitch": 68,
                        "velocity": 80,
                        "start_tick": 14,
                        "end_tick": 20,
                    },
                ],
                clip1,
            ),
        )
        # Note that start before the clip start are excluded.
        assert_notes_are_equal(
            list(clip2.get_notes()),
            create_test_notes(
                [
                    {
                        "pitch": 68,
                        "velocity": 80,
                        "start_tick": 24,
                        "end_tick": 35,
                        "id": 3
                    },
                    {
                        "pitch": 66,
                        "velocity": 80,
                        "start_tick": 25,
                        "end_tick": 30,
                        "id": 1
                    },
                ],
                clip2,
            ),
        )
        # Note are sorted and all notes that are within clip are returned.
        assert_notes_are_equal(
            list(clip3.get_notes()),
            create_test_notes(
                [
                    {
                        "pitch": 67,
                        "velocity": 80,
                        "start_tick": 40,
                        "end_tick": 45,
                        "id": 3
                    },
                    {
                        "pitch": 69,
                        "velocity": 80,
                        "start_tick": 45,
                        "end_tick": 50,
                        "id": 2
                    },
                    {
                        "pitch": 71,
                        "velocity": 80,
                        "start_tick": 55,
                        "end_tick": 65,
                        "id": 1
                    },
                ],
                clip3,
            ),
        )

    def test_clear_notes(self):
        track = self.song.get_track_at(0)
        clip1 = track.get_clip_at(0)
        self.assertGreater(clip1.get_raw_note_count(), 0)
        clip1.clear_notes()
        self.assertEqual(clip1.get_raw_note_count(), 0)


class TestCreateClip(BaseTestCase):
    def test_create_clip_returns_correct_reference(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip3 = self.song.get_track_at(0).create_midi_clip(
            clip_start_tick=999, clip_end_tick=1000)
        self.assertEqual(track.get_clip_at(3).get_clip_end_tick(), 1000)
        self.assertEqual(clip3.get_clip_end_tick(), 1000)
        clip3.adjust_clip_right(1200)
        self.assertEqual(track.get_clip_at(3).get_clip_end_tick(), 1200)
        self.assertEqual(clip3.get_clip_end_tick(), 1200)


class TestCreateNotes(BaseTestCase):
    def test_reject_creating_note_with_invalid_pitch(self):
        clip1 = self.song.get_track_at(0).get_clip_at(0)
        self.assertEqual(clip1.get_raw_note_count(), 3)
        assert_notes_are_equal(
            list(clip1.get_raw_notes()),
            create_test_notes(
                [
                    {
                        "pitch": 64,
                        "velocity": 80,
                        "start_tick": 0,
                        "end_tick": 10,
                        "id": 1
                    },
                    {
                        "pitch": 68,
                        "velocity": 80,
                        "start_tick": 14,
                        "end_tick": 20,
                        "id": 3
                    },
                    {
                        "pitch": 66,
                        "velocity": 80,
                        "start_tick": 15,
                        "end_tick": 20,
                        "id": 2
                    },
                ],
                clip1,
            ),
        )

        # Create notes with invalid pitch.
        created_note1 = clip1.create_note(
            pitch=128,
            velocity=80,
            start_tick=2,
            end_tick=8,
        )
        self.assertIsNone(created_note1)
        created_note2 = clip1.create_note(
            pitch=-1,
            velocity=80,
            start_tick=2,
            end_tick=8,
        )
        self.assertIsNone(created_note2)
        self.assertEqual(clip1.get_raw_note_count(), 3)
        assert_notes_are_equal(
            list(clip1.get_raw_notes()),
            create_test_notes(
                [
                    {
                        "pitch": 64,
                        "velocity": 80,
                        "start_tick": 0,
                        "end_tick": 10,
                        "id": 1
                    },
                    {
                        "pitch": 68,
                        "velocity": 80,
                        "start_tick": 14,
                        "end_tick": 20,
                        "id": 3
                    },
                    {
                        "pitch": 66,
                        "velocity": 80,
                        "start_tick": 15,
                        "end_tick": 20,
                        "id": 2
                    },
                ],
                clip1,
            ),
        )

    def test_reject_creating_note_with_invalid_velocity(self):
        clip1 = self.song.get_track_at(0).get_clip_at(0)
        self.assertEqual(clip1.get_raw_note_count(), 3)
        assert_notes_are_equal(
            list(clip1.get_raw_notes()),
            create_test_notes(
                [
                    {
                        "pitch": 64,
                        "velocity": 80,
                        "start_tick": 0,
                        "end_tick": 10,
                        "id": 1
                    },
                    {
                        "pitch": 68,
                        "velocity": 80,
                        "start_tick": 14,
                        "end_tick": 20,
                        "id": 3
                    },
                    {
                        "pitch": 66,
                        "velocity": 80,
                        "start_tick": 15,
                        "end_tick": 20,
                        "id": 2
                    },
                ],
                clip1,
            ),
        )

        # Create notes with invalid pitch.
        created_note1 = clip1.create_note(
            pitch=58,
            velocity=128,
            start_tick=2,
            end_tick=8,
        )
        self.assertIsNone(created_note1)
        created_note2 = clip1.create_note(
            pitch=56,
            velocity=-1,
            start_tick=2,
            end_tick=8,
        )
        self.assertIsNone(created_note2)
        self.assertEqual(clip1.get_raw_note_count(), 3)
        assert_notes_are_equal(
            list(clip1.get_raw_notes()),
            create_test_notes(
                [
                    {
                        "pitch": 64,
                        "velocity": 80,
                        "start_tick": 0,
                        "end_tick": 10,
                        "id": 1
                    },
                    {
                        "pitch": 68,
                        "velocity": 80,
                        "start_tick": 14,
                        "end_tick": 20,
                        "id": 3
                    },
                    {
                        "pitch": 66,
                        "velocity": 80,
                        "start_tick": 15,
                        "end_tick": 20,
                        "id": 2
                    },
                ],
                clip1,
            ),
        )

    def test_reject_creating_note_with_invalid_range(self):
        clip1 = self.song.get_track_at(0).get_clip_at(0)
        self.assertEqual(clip1.get_raw_note_count(), 3)
        assert_notes_are_equal(
            list(clip1.get_raw_notes()),
            create_test_notes(
                [
                    {
                        "pitch": 64,
                        "velocity": 80,
                        "start_tick": 0,
                        "end_tick": 10,
                        "id": 1
                    },
                    {
                        "pitch": 68,
                        "velocity": 80,
                        "start_tick": 14,
                        "end_tick": 20,
                        "id": 3
                    },
                    {
                        "pitch": 66,
                        "velocity": 80,
                        "start_tick": 15,
                        "end_tick": 20,
                        "id": 2
                    },
                ],
                clip1,
            ),
        )

        # Create notes with invalid pitch.
        created_note1 = clip1.create_note(
            pitch=58,
            velocity=80,
            start_tick=-20,
            end_tick=-1,
        )
        self.assertIsNone(created_note1)
        created_note2 = clip1.create_note(
            pitch=56,
            velocity=80,
            start_tick=9,
            end_tick=8,
        )
        self.assertIsNone(created_note2)
        self.assertEqual(clip1.get_raw_note_count(), 3)
        assert_notes_are_equal(
            list(clip1.get_raw_notes()),
            create_test_notes(
                [
                    {
                        "pitch": 64,
                        "velocity": 80,
                        "start_tick": 0,
                        "end_tick": 10,
                        "id": 1
                    },
                    {
                        "pitch": 68,
                        "velocity": 80,
                        "start_tick": 14,
                        "end_tick": 20,
                        "id": 3
                    },
                    {
                        "pitch": 66,
                        "velocity": 80,
                        "start_tick": 15,
                        "end_tick": 20,
                        "id": 2
                    },
                ],
                clip1,
            ),
        )

    def test_create_note_with_valid_params(self):
        clip1 = self.song.get_track_at(0).get_clip_at(0)
        self.assertEqual(clip1.get_raw_note_count(), 3)
        assert_notes_are_equal(
            list(clip1.get_raw_notes()),
            create_test_notes(
                [
                    {
                        "pitch": 64,
                        "velocity": 80,
                        "start_tick": 0,
                        "end_tick": 10,
                        "id": 1
                    },
                    {
                        "pitch": 68,
                        "velocity": 80,
                        "start_tick": 14,
                        "end_tick": 20,
                        "id": 3
                    },
                    {
                        "pitch": 66,
                        "velocity": 80,
                        "start_tick": 15,
                        "end_tick": 20,
                        "id": 2
                    },
                ],
                clip1,
            ),
        )

        # Create notes with invalid pitch.
        created_note1 = clip1.create_note(
            pitch=58,
            velocity=80,
            start_tick=12,
            end_tick=14,
        )
        assert_notes_are_equal(
            [created_note1],
            create_test_notes(
                [
                    {
                        "pitch": 58,
                        "velocity": 80,
                        "start_tick": 12,
                        "end_tick": 14,
                        "id": 4
                    },
                ],
                clip1,
            ),
        )
        self.assertEqual(clip1.get_raw_note_count(), 4)
        assert_notes_are_equal(
            list(clip1.get_raw_notes()),
            create_test_notes(
                [
                    {
                        "pitch": 64,
                        "velocity": 80,
                        "start_tick": 0,
                        "end_tick": 10,
                        "id": 1
                    },
                    {
                        "pitch": 58,
                        "velocity": 80,
                        "start_tick": 12,
                        "end_tick": 14,
                        "id": 4
                    },
                    {
                        "pitch": 68,
                        "velocity": 80,
                        "start_tick": 14,
                        "end_tick": 20,
                        "id": 3
                    },
                    {
                        "pitch": 66,
                        "velocity": 80,
                        "start_tick": 15,
                        "end_tick": 20,
                        "id": 2
                    },
                ],
                clip1,
            ),
        )

    def test_create_note_returns_correct_reference(self):
        clip1 = self.song.get_track_at(0).get_clip_at(0)
        self.assertEqual(clip1.get_raw_note_count(), 3)

        # Create notes with invalid pitch.
        created_note1 = clip1.create_note(
            pitch=127,
            velocity=80,
            start_tick=999,
            end_tick=1000,
        )

        self.assertEqual(clip1.get_raw_note_at(3).get_pitch(), 127)
        self.assertEqual(created_note1.get_pitch(), 127)  # type:ignore
        created_note1.set_pitch(1)  # type:ignore
        self.assertEqual(clip1.get_raw_note_at(3).get_pitch(), 1)
        self.assertEqual(created_note1.get_pitch(), 1)  # type:ignore


class TestMoveClips(BaseTestCase):
    def test_move_clips_no_overlapping(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip1.move_clip_to(70, move_associated_track_automation_points=False)
        self.assert_clip_range(clip1, 70, 85)
        self.assert_clip_range(clip2, 21, 30)
        self.assert_clip_range(clip3, 40, 65)

    def test_move_clips_overlaps_on_the_left_side(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip1.move_clip_to(10, move_associated_track_automation_points=False)
        self.assert_clip_range(clip1, 10, 25)
        self.assert_clip_range(clip2, 26, 30)
        self.assert_clip_range(clip3, 40, 65)

    def test_move_clips_overlaps_on_the_right_side(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip1.move_clip_to(60, move_associated_track_automation_points=False)
        self.assert_clip_range(clip1, 60, 75)
        self.assert_clip_range(clip2, 21, 30)
        self.assert_clip_range(clip3, 40, 59)

    def test_move_clips_overlaps_on_the_left_and_right_side(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip1.move_clip_to(28, move_associated_track_automation_points=False)
        self.assert_clip_range(clip1, 28, 43)
        self.assert_clip_range(clip2, 21, 27)
        self.assert_clip_range(clip3, 44, 65)

    def test_move_clips_overlaps_in_the_middle(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip2.move_clip_to(42, move_associated_track_automation_points=False)
        self.assertEqual(track.get_clip_count(), 4)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 40, 41)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 42, 51)
        clip4 = track.get_clip_at(3)
        self.assert_clip_range(clip4, 52, 65)

    def test_move_clips_delete_out_of_range_clip(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip1.move_clip(-99999, move_associated_track_automation_points=False)
        self.assertEqual(track.get_clip_count(), 2)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 21, 30)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 40, 65)

    def test_move_clips_move_associated_track_automation_points(self):
        # TODO: Add tests
        pass


class TestAdjustClipRanges(BaseTestCase):
    def test_adjust_clip_left_no_overlapping(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip3.adjust_clip_left(35)
        self.assert_clip_range(clip1, 0, 15)
        self.assert_clip_range(clip2, 21, 30)
        self.assert_clip_range(clip3, 35, 65)

    def test_adjust_clip_left_overlaps_on_the_right_side(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip3.adjust_clip_left(25)
        self.assert_clip_range(clip1, 0, 15)
        self.assert_clip_range(clip2, 21, 24)
        self.assert_clip_range(clip3, 25, 65)

    def test_adjust_clip_left_overlaps_on_the_whole_clip_and_on_the_right(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip3.adjust_clip_left(10)
        self.assertEqual(track.get_clip_count(), 2)
        clip1 = track.get_clip_at(0)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip1, 0, 9)
        self.assert_clip_range(clip2, 10, 65)

    def test_adjust_clip_left_overlaps_two_whole_clips(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip3.adjust_clip_left(-10)
        self.assertEqual(track.get_clip_count(), 1)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 65)

    def test_adjust_clip_right_no_overlapping(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip1.adjust_clip_right(18)
        self.assert_clip_range(clip1, 0, 18)
        self.assert_clip_range(clip2, 21, 30)
        self.assert_clip_range(clip3, 40, 65)

    def test_adjust_clip_right_overlaps_on_the_left_side(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip1.adjust_clip_right(25)
        self.assert_clip_range(clip1, 0, 25)
        self.assert_clip_range(clip2, 26, 30)
        self.assert_clip_range(clip3, 40, 65)

    def test_adjust_clip_right_overlaps_on_the_whole_clip_and_on_the_left(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip1.adjust_clip_right(50)
        self.assertEqual(track.get_clip_count(), 2)
        clip1 = track.get_clip_at(0)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip1, 0, 50)
        self.assert_clip_range(clip2, 51, 65)

    def test_adjust_clip_right_overlaps_two_whole_clips(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip1.adjust_clip_right(1000)
        self.assertEqual(track.get_clip_count(), 1)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 1000)

    def test_adjust_clip_range_no_overlapping(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip2.adjust_clip_left(18)
        clip2.adjust_clip_right(35)
        self.assert_clip_range(clip1, 0, 15)
        self.assert_clip_range(clip2, 18, 35)
        self.assert_clip_range(clip3, 40, 65)

    def test_adjust_clip_range_overlaps_on_the_left_side(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip2.adjust_clip_left(10)
        clip2.adjust_clip_right(35)
        self.assert_clip_range(clip1, 0, 9)
        self.assert_clip_range(clip2, 10, 35)
        self.assert_clip_range(clip3, 40, 65)

    def test_adjust_clip_range_overlaps_on_the_whole_clip_to_the_left(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip2.adjust_clip_left(0)
        clip2.adjust_clip_right(35)
        self.assertEqual(track.get_clip_count(), 2)
        clip1 = track.get_clip_at(0)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip1, 0, 35)
        self.assert_clip_range(clip2, 40, 65)

    def test_adjust_clip_range_overlaps_on_the_right_side(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip2.adjust_clip_left(21)
        clip2.adjust_clip_right(45)
        self.assert_clip_range(clip1, 0, 15)
        self.assert_clip_range(clip2, 21, 45)
        self.assert_clip_range(clip3, 46, 65)

    def test_adjust_clip_range_overlaps_on_the_whole_clip_to_the_right(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip2.adjust_clip_left(21)
        clip2.adjust_clip_right(1000)
        self.assertEqual(track.get_clip_count(), 2)
        clip1 = track.get_clip_at(0)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip1, 0, 15)
        self.assert_clip_range(clip2, 21, 1000)

    def test_adjust_clip_range_overlaps_on_the_left_and_right_side(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip2.adjust_clip_left(10)
        clip2.adjust_clip_right(45)
        self.assert_clip_range(clip1, 0, 9)
        self.assert_clip_range(clip2, 10, 45)
        self.assert_clip_range(clip3, 46, 65)

    def test_adjust_clip_range_overlaps_two_whole_clips(self):
        track = self.song.get_track_at(0)
        self.assertEqual(track.get_clip_count(), 3)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 40, 65)
        clip2.adjust_clip_left(0)
        clip2.adjust_clip_right(1000)
        self.assertEqual(track.get_clip_count(), 1)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 1000)


class TestInsertClip(BaseTestCase):
    def test_insert_clip_no_overlapping(self):
        track = self.song.get_track_at(0)
        track.create_midi_clip(
            clip_start_tick=32,
            clip_end_tick=38,
        )
        self.assertEqual(track.get_clip_count(), 4)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 30)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 32, 38)
        clip4 = track.get_clip_at(3)
        self.assert_clip_range(clip4, 40, 65)

    def test_insert_clip_overalaps_with_left_and_right(self):
        track = self.song.get_track_at(0)
        track.create_midi_clip(
            clip_start_tick=25,
            clip_end_tick=45,
        )
        self.assertEqual(track.get_clip_count(), 4)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 24)
        clip3 = track.get_clip_at(2)
        self.assert_clip_range(clip3, 25, 45)
        clip4 = track.get_clip_at(3)
        self.assert_clip_range(clip4, 46, 65)

    def test_insert_clip_overlaps_with_whole_clips(self):
        track = self.song.get_track_at(0)
        track.create_midi_clip(
            clip_start_tick=21,
            clip_end_tick=65,
        )
        self.assertEqual(track.get_clip_count(), 2)
        clip1 = track.get_clip_at(0)
        self.assert_clip_range(clip1, 0, 15)
        clip2 = track.get_clip_at(1)
        self.assert_clip_range(clip2, 21, 65)


class TestDeleteClip(BaseTestCase):
    # TODO: Add tests
    pass


if __name__ == '__main__':
    unittest.main()



from tuneflow_py.descriptors.param import ParamDescriptor
from typing import Any, Dict
from tuneflow_py.models.song import Song


class TuneflowPlugin:
    '''
    The base class of a plugin.

    All plugins should be a sub-class of this plugin in order to run in the pipeline.
    '''

    @staticmethod
    def provider_id() -> str:
        """
        The unique id to identify the plugin provider.

        For example:
        `friday-core`, `some-corp`, etc.
        """
        raise Exception("provider_id must be overwritten.")

    @staticmethod
    def plugin_id() -> str:
        '''
        The unique plugin id under the provider's namespace.

        For example:
        `melody-extractor`, `tune-analyzer`, etc.
        '''
        raise Exception("plugin_id must be overwritten.")

    @staticmethod
    def params(song: Song) -> Dict[str, ParamDescriptor]:
        '''
        Specify params to get from user input.

        Param input widgets will be displayed on the UI, and the inputs will be collected and fed into @run method.

        You can also optionally use `song` and `read_apis` to adjust your params based on the current snapshot.

        If you don't need any param, return `{}`;
        '''
        return {}

    @staticmethod
    def run(song: Song, params: Dict[str, Any]):
        '''
        The main logic here.

        @param song The song that is being processed. You can directly modify the song
                by calling its methods.
        @param params The results collected from user input specified by the `params` method.
        '''
        pass

    # =====================================
    # NO OVERWRITE BELOW
    # =====================================

    @staticmethod
    def _get_default_params(param_config):
        param_result = {}
        for key in param_config.keys():
            param_descriptor = param_config[key]
            param_result[key] = param_descriptor["defaultValue"]
        return param_result



from __future__ import annotations
from tuneflow_py.descriptors.text import LabelText
from tuneflow_py.descriptors.common import RealNumber
from enum import Enum
from typing import Any, List, Dict
from typing_extensions import Literal, TypedDict, Required, NotRequired
from tuneflow_py.models.track import TrackType


class WidgetType(Enum):
    Slider = 1
    Input = 2
    # A widget that selects one track.
    TrackSelector = 3
    # A widget that selects a pitch.
    Pitch = 4
    TrackPitchSelector = 5
    InstrumentSelector = 6
    Select = 7
    Switch = 8
    InputNumber = 9
    MultiTrackSelector = 10
    # Does not render a widget.
    NoWidget = 11
    FileSelector = 12
    MultiSourceAudioSelector = 13
    AudioRecorder = 14
    SelectList = 15
    # Read-only table of descriptions.
    Descriptions = 16
    TextArea = 17


SliderMarkerConfig = Dict[RealNumber, LabelText]


class SliderWidgetConfig(TypedDict):
    minValue: Required[RealNumber]
    maxValue: Required[RealNumber]
    step: Required[RealNumber]
    unit: NotRequired[str]
    markers: NotRequired[SliderMarkerConfig]


class InputWidgetConfig(TypedDict):
    minValue: Required[RealNumber]
    maxValue: Required[RealNumber]
    step: Required[RealNumber]


class TrackSelectorWidgetConfig(TypedDict):
    alwaysShowTrackInfo: NotRequired[bool]
    '''  Whether to always show the track info. Default to false. '''

    allowedTrackTypes: NotRequired[List[TrackType]]  # type: ignore
    ''' If specified, only the allowed types of tracks can be selected. '''


class PitchWidgetConfig(TypedDict):
    minAllowedPitch: NotRequired[RealNumber]
    maxAllowedPitch: NotRequired[RealNumber]


class TrackPitchSelectorWidgetConfig(TypedDict):
    trackSelectorConfig: Required[TrackSelectorWidgetConfig]
    pitchSelectorConfig: Required[PitchWidgetConfig]


class InstrumentSelectorWidgetConfig(TypedDict):
    disabledPrograms: NotRequired[List[RealNumber]]
    ''' Not supported yet. '''


class SelectWidgetOption(TypedDict):
    label: Required[LabelText]
    value: Required[Any]


class SelectWidgetConfig(TypedDict):
    options: Required[List[SelectWidgetOption]]

    allowSearch: NotRequired[bool]
    ''' Whether to show search box. Default to false. '''

    placeholder: Required[LabelText]

    virtualListProps: NotRequired[Any]
    ''' https://arco.design/vue/component/select#virtual-list '''

    populateOptionsWithGeneratableStyles: NotRequired[bool]
    '''
    Whether to populate the options with the styles that
    TuneFlow can generate.
    '''

    populateOptionsWithGeneratableTempos: NotRequired[bool]
    '''
    Whether to populate the options with the tempo settings that
    TuneFlow can generate with.
    '''


class SelectListWidgetConfig(TypedDict):
    options: Required[List[SelectWidgetOption]]
    maxHeight: NotRequired[RealNumber]
    size: NotRequired[str]
    virtualListProps: NotRequired[Any]
    ''' https://arco.design/vue/component/list '''

    allowSearch: NotRequired[bool]


class SwitchWidgetConfig(TypedDict):
    type: NotRequired[Literal['circle', 'round', 'line']]
    '''
    'circle' | 'round' | 'line'
    
    https://arco.design/vue/component/switch
    '''


class InputNumberWidgetConfig(TypedDict):
    minValue: Required[RealNumber]
    maxValue: Required[RealNumber]
    step: Required[RealNumber]


class FileSelectorWidgetConfig(TypedDict):
    allowedExtensions: Required[List[str]]
    ''' The extensions (without ".") that are allowed to choose. '''

    selectDirectory: NotRequired[str]
    '''
    Whether to select a directory instead of a file.
    Default to false.
    '''

    placeholder: NotRequired[LabelText]
    ''' Custom placeholder text. '''

    selectLocalFile: NotRequired[bool]
    ''' If true, selects local system files. '''


AudioSourceType = Literal['file', 'audioTrack', 'record']


class MultiSourceAudioSelectorWidgetConfig(TypedDict):
    allowedSources: NotRequired[List[AudioSourceType]]
    ''' Default to allow all audio sources. '''


class MultiSourceAudioSelectorResult(TypedDict):
    sourceType: Required[AudioSourceType]

    audioInfo: Required[Any]
    '''
    Result type will be:
    * `File` if `sourceType` is 'file'
    * trackId if `sourceType` is 'audioTrack'
    * `AudioBuffer` if `sourceType` is 'record'
    '''


class DescriptionData(TypedDict):
    label: Required[LabelText]
    value: Required[str]
    span: NotRequired[RealNumber]


class DescriptionsWidgetConfig(TypedDict):
    size: Required[Literal['mini', 'small', 'medium', 'large']]
    column: Required[int]
    data: Required[List[DescriptionData]]


class TextAreaWidgetConfig(TypedDict):
    placeholder: NotRequired[str]
    maxLength: NotRequired[int]
    allowClear: NotRequired[bool]
    autoSize: NotRequired[bool]


class WidgetDescriptor(TypedDict):
    type: Required[int]  # WidgetType
    config: NotRequired[SliderWidgetConfig
                        | InputWidgetConfig
                        | SelectWidgetConfig
                        | TrackSelectorWidgetConfig
                        | PitchWidgetConfig
                        | TrackPitchSelectorWidgetConfig
                        | InstrumentSelectorWidgetConfig
                        | SwitchWidgetConfig
                        | InputNumberWidgetConfig
                        | FileSelectorWidgetConfig
                        | SelectListWidgetConfig
                        | MultiSourceAudioSelectorWidgetConfig
                        | DescriptionsWidgetConfig
                        | TextAreaWidgetConfig]



from tuneflow_py.utils import greater_than, greater_equal, lower_than, lower_equal
import unittest


class TestBinarySearch(unittest.TestCase):
    def test_greater_equal(self):
        self.assertEqual(greater_equal([1, 2, 3, 3, 4, 5], 0), 0)
        self.assertEqual(greater_equal([1, 2, 3, 3, 4, 5], 1), 0)
        self.assertEqual(greater_equal([1, 2, 3, 3, 4, 5], 3), 2)
        self.assertEqual(greater_equal([{"val": 1}, {"val": 2}, {"val": 3}, {
                         "val": 3}, {"val": 4}, {"val": 5}], {"val": 3}, key=lambda x: x["val"]), 2)
        self.assertEqual(greater_equal([1, 2, 3, 3, 4, 5], 6), 6)

    def test_greater_than(self):
        self.assertEqual(greater_than([1, 2, 3, 3, 4, 5], 0), 0)
        self.assertEqual(greater_than([1, 2, 3, 3, 4, 5], 1), 1)
        self.assertEqual(greater_than([1, 2, 3, 3, 4, 5], 3), 4)
        self.assertEqual(greater_than([{"val": 1}, {"val": 2}, {"val": 3}, {
                         "val": 3}, {"val": 4}, {"val": 5}], {"val": 3}, key=lambda x: x["val"]), 4)
        self.assertEqual(greater_than([1, 2, 3, 3, 4, 5], 6), 6)

    def test_lower_equal(self):
        self.assertEqual(lower_equal([1, 2, 3, 3, 4, 5], 0), -1)
        self.assertEqual(lower_equal([1, 2, 3, 3, 4, 5], 1), 0)
        self.assertEqual(lower_equal([1, 2, 3, 3, 4, 5], 3), 3)
        self.assertEqual(lower_equal([{"val": 1}, {"val": 2}, {"val": 3}, {
                         "val": 3}, {"val": 4}, {"val": 5}], {"val": 3}, key=lambda x: x["val"]), 3)
        self.assertEqual(lower_equal([1, 2, 3, 3, 4, 5], 6), 5)

    def test_lower_than(self):
        self.assertEqual(lower_than([1, 2, 3, 3, 4, 5], 0), -1)
        self.assertEqual(lower_than([1, 2, 3, 3, 4, 5], 1), -1)
        self.assertEqual(lower_than([1, 2, 3, 3, 4, 5], 3), 1)
        self.assertEqual(lower_than([{"val": 1}, {"val": 2}, {"val": 3}, {
                         "val": 3}, {"val": 4}, {"val": 5}], {"val": 3}, key=lambda x: x["val"]), 1)
        self.assertEqual(lower_than([1, 2, 3, 3, 4, 5], 6), 5)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from tuneflow_py.models.protos import song_pb2
from tuneflow_py.utils import greater_equal, greater_than, lower_than
from types import SimpleNamespace
from typing import List
from typing_extensions import TypedDict, Required, Any

AutomationTargetType = song_pb2.AutomationTarget.TargetType


class AutomationTarget:
    def __init__(
            self, type: AutomationTargetType | None = None, plugin_instance_id: str | None = None, param_id: str | None = None,
            proto: song_pb2.AutomationTarget | None = None) -> None:
        if proto is not None:
            self._proto = proto
        else:
            self._proto = song_pb2.AutomationTarget(
                type=type, audio_plugin_id=plugin_instance_id, param_id=param_id)

    def get_type(self):
        return self._proto.type

    def set_type(self, type: AutomationTargetType):
        self._proto.type = type

    def get_plugin_instance_id(self):
        return self._proto.audio_plugin_id

    def set_plugin_instance_id(self, plugin_instance_id: str | None = None):
        if plugin_instance_id is None:
            self._proto.ClearField("audio_plugin_id")
        else:
            self._proto.audio_plugin_id = plugin_instance_id

    def get_param_id(self):
        return self._proto.param_id

    def set_param_id(self, param_id: str | None = None):
        if param_id is None:
            self._proto.ClearField("param_id")
        else:
            self._proto.param_id = param_id

    def equals(self, target: AutomationTarget):
        return AutomationTarget.are_automation_targets_equal(
            self.get_type(),
            target.get_type(),
            self.get_plugin_instance_id(),
            target.get_plugin_instance_id(),
            self.get_param_id(),
            target.get_param_id(),
        )

    def clone(self):
        new_proto = song_pb2.AutomationTarget()
        new_proto.CopyFrom(self._proto)
        return AutomationTarget(proto=new_proto)

    def to_tf_automation_target_id(self):
        '''
        Gets a unique string id that identifies this target type.
        '''
        return AutomationTarget.encode_automation_target(
            self.get_type(),
            self.get_plugin_instance_id(),
            self.get_param_id())

    def __repr__(self) -> str:
        return str(self._proto)

    @staticmethod
    def encode_automation_target(
        target_type: AutomationTargetType,
        plugin_instance_id: str | None,
        param_id: str | None,
    ):
        if target_type == AutomationTargetType.AUDIO_PLUGIN:
            return f'{target_type}^^{plugin_instance_id}^^{param_id}'

        return f'{target_type}'

    @staticmethod
    def decode_automation_target(encoded_target: str):
        parts = encoded_target.split('^^')
        if len(parts) == 0:
            raise Exception(f'Invalid automation target id: {encoded_target}')

        type = int(parts[0])
        if len(parts) > 2:
            AutomationTarget(type, parts[1], parts[2])

        return AutomationTarget(type)

    @staticmethod
    def are_automation_targets_equal(
        target_type1: AutomationTargetType,
        target_type2: AutomationTargetType,
        plugin_instance_id1: str | None,
        plugin_instance_id2: str | None,
        param_id1: str | None,
        param_id2: str | None
    ):
        return (
            AutomationTarget.encode_automation_target(target_type1, plugin_instance_id1, param_id1) ==
            AutomationTarget.encode_automation_target(target_type2, plugin_instance_id2, param_id2)
        )


class AutomationPoint(TypedDict):
    '''
    A single point in an automation curve.
    '''
    id: Required[int]
    tick:  Required[int]
    value:  Required[float]


class AutomationValue:
    '''
    The points and settings of an automation param.
    '''

    def __init__(self, proto: song_pb2.AutomationValue | None = None):
        if proto is not None:
            self._proto = proto
        else:
            self._proto = song_pb2.AutomationValue()
            self._proto.disabled = False
        self._next_point_id = None

    def get_disabled(self):
        return self._proto.disabled

    def set_disabled(self, is_disabled: bool):
        self._proto.disabled = is_disabled

    def get_points(self) -> List[AutomationPoint]:
        return self._proto.points

    def get_points_in_range(self, start_tick: int, end_tick: int):
        target_point = SimpleNamespace()
        target_point.tick = start_tick
        start_index = greater_equal(
            self._proto.points,
            target_point,
            lambda x: x.tick,
        )
        results = []
        for i in range(start_index, len(self._proto.points)):
            point = self._proto.points[i]
            if (point.tick > end_tick):
                break
            results.append(point)

        return results

    def add_point(self, tick: int, value: float, overwrite=False):
        '''
        @param overwrite Whether to overwrite the points at the insert tick.
        '''
        new_point = song_pb2.AutomationValue.ParamValue(
            tick=tick,
            value=max(0, min(1, value)),
            id=self._get_next_point_id(),
        )
        return self._ordered_insert_point(self._proto.points, new_point, overwrite)

    def remove_points(self, point_ids: List[int]):
        '''
        Remove points that match the given ids.
        '''
        id_set = set(point_ids)
        for i in range(len(self._proto.points)-1, -1, -1):
            point = self._proto.points[i]
            if point.id in id_set:
                self._proto.points.pop(i)

    def remove_points_in_range(self, start_tick: int, end_tick: int):
        '''
        Removes all points within the given time range.
        @param start_tick Inclusive
        @param end_tick Inclusive
        '''
        target_point = SimpleNamespace()
        target_point.tick = start_tick
        start_index = greater_equal(
            self._proto.points,
            target_point,
            lambda x: x.tick,
        )
        if start_index >= len(self._proto.points):
            return

        end_index = start_index
        while (end_index + 1 < len(self._proto.points) and self._proto.points[end_index + 1].tick <= end_tick):
            end_index += 1
        del self._proto.points[start_index:end_index+1]

    def move_points_in_range(
        self,
        start_tick: int,
        end_tick: int,
        offset_tick: int,
        offset_value: float,
        overwrite_values_in_drag_area=True,
    ):
        points = self.get_points_in_range(start_tick, end_tick)
        self.move_points(
            [point.id for point in points],
            offset_tick,
            offset_value,
            overwrite_values_in_drag_area,
        )

    def move_all_points(self, offset_tick: int, offset_value: float, overwrite_values_in_drag_area=True):
        self.move_points(
            [point.id for point in self._proto.points],
            offset_tick,
            offset_value,
            overwrite_values_in_drag_area,
        )

    def move_points(
            self,
        point_ids: List[int],
        offset_tick: int,
        offset_value: float,
        overwrite_values_in_drag_area=True,
    ):
        '''
        @param overwrite_values_in_drag_area If true, all values in between the moved points' old and new indexes will be removed.
        '''
        if (len(point_ids) == 0):
            return

        point_id_set = set(point_ids)
        drag_area_left_index = None
        drag_area_right_index = None
        selected_points = []
        for i in range(len(self._proto.points)):
            point = self._proto.points[i]
            if point.id not in point_id_set:
                continue

            selected_points.append(point)
            if drag_area_left_index is None:
                drag_area_left_index = i

            drag_area_right_index = i

        if drag_area_left_index is None or drag_area_right_index is None:
            # None of the given points are not in the automation.
            return

        if overwrite_values_in_drag_area:
            # Remove values in drag area.
            if (offset_tick < 0):
                # Move left, remove values to the left.
                selected_points_left_after_move = max(
                    0,
                    self._proto.points[drag_area_left_index].tick + offset_tick,
                )
                target_point = SimpleNamespace()
                target_point.tick = selected_points_left_after_move
                start_remove_index = greater_than(
                    self._proto.points,
                    target_point,
                    lambda x: x.tick,
                )
                if (start_remove_index < drag_area_left_index):
                    del self._proto.points[start_remove_index:drag_area_left_index]
            elif (offset_tick > 0):
                # Move right, remove values to the right.
                selected_points_right_after_move = self._proto.points[drag_area_right_index].tick + offset_tick
                target_point = SimpleNamespace()
                target_point.tick = selected_points_right_after_move
                end_remove_index = lower_than(
                    self._proto.points,
                    target_point,
                    lambda x: x.tick,
                )
                if (end_remove_index > drag_area_right_index):
                    del self._proto.points[drag_area_right_index + 1:end_remove_index+1]

        for point in selected_points:
            point.tick = max(0, point.tick + offset_tick)
            point.value = max(0, min(1, point.value + offset_value))

        # Maintain the order of points.
        if (abs(offset_tick) > 0):
            self._proto.points.sort(key=lambda x: x.tick)

    def clone(self):
        new_proto = song_pb2.AutomationValue()
        new_proto.CopyFrom(self._proto)
        return AutomationValue(proto=new_proto)

    def _get_next_point_id(self):
        if self._next_point_id is None:
            if len(self._proto.points) == 0:
                self._next_point_id = 1
            else:
                self._next_point_id = max([point.id for point in self._proto.points]) + 1

        point_id = self._next_point_id
        if (self._next_point_id >= 2147483647):
            self._next_point_id = 1
        else:
            self._next_point_id += 1

        return point_id

    def __eq__(self, __value: AutomationValue) -> bool:
        return self._proto == __value._proto

    @staticmethod
    def _ordered_insert_point(
        points: Any,
        new_point: song_pb2.AutomationValue.ParamValue,
        overwrite=False,
    ):
        '''
        Be careful that only a copy of the new_point will be inserted, not the original object.

        So you need to update your reference to the new_point after inserting it.

        @return The copy of the original point that is actually inserted into points.
        '''
        insert_index = greater_equal(
            points,
            new_point,
            lambda x: x.tick
        )

        while (overwrite and insert_index >= 0 and insert_index < len(points) and points[insert_index].tick == new_point.tick):
            # Remove the points at the insert tick.
            del points[insert_index]
        points.insert(insert_index, new_point)
        return points[insert_index]


class AutomationData:
    '''
    All automation data of one entity (such as a track).
    *
    * Each `AutomationData` consists of several automation targets(`AutomationTarget`) and
    * the values(`AutomationValue`) store in unique targets.
    *
    * Note that there can be duplicate targets, but targets of the same type write to
    * and read from the same automation value.
    *
    * For example, there can be multiple Volume targets, each Volume target corresponds to
    * the same automation value.
    '''

    def __init__(self, proto: song_pb2.AutomationData | None = None):
        if proto is not None:
            self._proto = proto
        else:
            self._proto = song_pb2.AutomationData()

    def get_automation_targets(self):
        '''
        All automation targets specified by the user.

        Note that the return of this method is a generator,
        you would need to convert it to a list to be able
        to access it with an index.
        '''
        for target in self._proto.targets:
            yield AutomationTarget(proto=target)

    def get_automation_target_values(self):
        '''
        Values of each unique automation target.
        '''
        raise Exception("get_automation_target_values not implemented yet")

    def get_or_create_automation_value_by_id(self, tf_automation_target_id: str) -> AutomationValue:
        '''
        @param tf_automation_target_id The targetId that can be retrieved from `AutomationTarget.prototype.toTfAutomationTargetId` or `AutomationTarget.encodeAutomationTarget`.
        @returns The automation value of the given target if exists, otherwise creates a new one and returns it.
        '''
        return AutomationValue(proto=self._proto.target_values[tf_automation_target_id])

    def get_automation_value_by_id(self, tf_automation_target_id: str):
        '''
        Gets or creates the automation points and settings of an automation target.
        @param tfAutomationTargetId The targetId that can be retrieved from `AutomationTarget.prototype.toTfAutomationTargetId` or `AutomationTarget.encodeAutomationTarget`.
        '''
        if tf_automation_target_id not in self._proto.target_values:
            return None
        return AutomationValue(proto=self._proto.target_values[tf_automation_target_id])

    def get_automation_value_by_target(self, target: AutomationTarget):
        tf_automation_target_id = target.to_tf_automation_target_id()
        return self.get_automation_value_by_id(tf_automation_target_id)

    def add_automation(self, target: AutomationTarget, index=0):
        '''
        Adds an automation target, if there was no such target, creates the automation value.
        '''
        if not isinstance(index, int):
            index = 0
        self._proto.targets.insert(index, target._proto)
        target._proto = self._proto.targets[index]
        tf_automation_target_id = target.to_tf_automation_target_id()
        self.get_or_create_automation_value_by_id(tf_automation_target_id)

    def remove_automation(self, target: AutomationTarget):
        '''
        Removes all automation targets of the given type and its automation value.
        '''
        # Remove targets.
        for i in range(len(self._proto.targets)-1, -1, -1):
            existing_target = AutomationTarget(proto=self._proto.targets[i])
            if (existing_target.equals(target)):
                del self._proto.targets[i]

        # Remove values.
        tf_automation_target_id = target.to_tf_automation_target_id()
        del self._proto.target_values[tf_automation_target_id]

    def remove_automation_of_plugin(self, plugin_instance_id: str):
        '''
        Remove all automations associated with a certain plugin.
        '''
        for i in range(len(self._proto.targets)-1, -1, -1):
            automation_target = AutomationTarget(proto=self._proto.targets[i])
            if (automation_target.get_plugin_instance_id() == plugin_instance_id):
                self.remove_automation(automation_target)
        for tf_automation_target_id in self._proto.target_values:
            automation_target = AutomationTarget.decode_automation_target(tf_automation_target_id)
            if (automation_target.get_plugin_instance_id() == plugin_instance_id):
                self.remove_automation(automation_target)

    def remove_all_points_within_range(self, start_tick: int, end_tick: int):
        '''
        @param start_tick Inclusive
        @param end_tick Inclusive
        '''
        for tf_automation_target_id in self._proto.target_values:
            automation_value = AutomationValue(proto=self._proto.targetValues[tf_automation_target_id])
            automation_value.remove_points_in_range(start_tick, end_tick)

    def move_all_points_within_range(
        self,
        start_tick: int,
        end_tick: int,
        offset_tick: int,
        offset_value: float,
    ):
        '''
        @param start_tick Inclusive
        @param end_tick Inclusive
        '''
        for tf_automation_target_id in self._proto.target_values:
            automation_value = AutomationValue(proto=self._proto.targetValues[tf_automation_target_id])
            automation_value.move_points_in_range(
                start_tick,
                end_tick,
                offset_tick,
                offset_value,
                overwrite_values_in_drag_area=False,
            )

    def clone(self):
        '''
        Creates a clone of this automation data.
        '''
        new_proto = song_pb2.AutomationData()
        new_proto.CopyFrom(self._proto)
        return AutomationData(proto=new_proto)



from __future__ import annotations
from base64 import b64encode, b64decode
from tuneflow_py.models.protos import song_pb2
from tuneflow_py.models.track import Track, TrackType, TrackOutputType
from tuneflow_py.models.marker import StructureMarker, StructureType
from tuneflow_py.models.clip import Clip, ClipType
from tuneflow_py.models.tempo import TempoEvent
from tuneflow_py.models.time_signature import TimeSignatureEvent
from tuneflow_py.models.automation import AutomationTarget, AutomationTargetType
from tuneflow_py.models.audio_plugin import AudioPlugin, decode_audio_plugin_tuneflow_id
from tuneflow_py.utils import db_to_volume_value, greater_equal, lower_than, lower_equal
from miditoolkit.midi import MidiFile, TempoChange as ToolkitTempoChange, TimeSignature as ToolkitTimeSignature, \
    Instrument, Note as ToolkitNote
from types import SimpleNamespace
from typing import List


class Song:
    def __init__(self, proto: song_pb2.Song | None = None) -> None:
        if proto is not None:
            self._proto = proto
        else:
            self._proto = song_pb2.Song()
            self._proto.PPQ = Song.get_default_resolution()
            self._proto.tempos.add(ticks=0, time=0, bpm=120)
            self._proto.time_signatures.add(ticks=0, numerator=4, denominator=4)
            self._proto.last_tick = 0
            self._proto.duration = 0
            self._proto.master_track.uuid = Track._generate_track_id()
            self._proto.master_track.type = TrackType.MASTER_TRACK
            self._proto.master_track.volume = db_to_volume_value(0.0)

    def get_last_tick(self):
        '''
        @returns End tick of the last note.
        '''
        if self.get_track_count() == 0:
            return 0
        return max([track.get_track_end_tick() for track in self.get_tracks()])

    def get_duration(self):
        return self.tick_to_seconds(self.get_last_tick())

    def get_track_count(self):
        return len(self._proto.tracks)

    def get_tracks(self):
        for track_proto in self._proto.tracks:
            yield Track(song=self, proto=track_proto)

    def get_track_by_id(self, track_id: str) -> Track | None:
        for track_proto in self._proto.tracks:
            if track_proto.uuid == track_id:
                return Track(song=self, proto=track_proto)
        return None

    def get_track_at(self, index):
        return Track(song=self, proto=self._proto.tracks[index])

    def get_track_index(self, track_id: str):
        '''
        Get the index of the track within the tracks list.
        Returns -1 if no track matches the track id.
        '''
        for index, track in enumerate(self.get_tracks()):
            if track.get_id() == track_id:
                return index
        return -1

    def remove_track(self, track_id: str):
        '''
        Removes a track from the song and returns it.
        '''
        track = self.get_track_by_id(track_id=track_id)
        if not track:
            return None

        for i in range(self.get_track_count() - 1, -1, -1):
            if self.get_track_at(i).get_id() == track_id:
                del self._proto.tracks[i]
        # Delete dependencies.
        for dep_track in self.get_tracks():
            track_output = dep_track.get_output()
            if track_output is not None and track_output.get_type() == TrackOutputType.TRACK_OUTPUT_TRACK and track_output.get_track_id() == track_id:
                dep_track.remove_output()
        return track

    def get_structures(self):
        return [StructureMarker(song=self, proto=structure_proto) for structure_proto in self._proto.structures]

    def get_structure_at_index(self, index: int):
        if index < 0 or index >= len(self._proto.structures):
            return None
        return StructureMarker(song=self, proto=self._proto.structures[index])

    def get_structure_at_tick(self, tick: int):
        def tick_to_structure_fn(tick):
            return StructureMarker(song=self, tick=tick, type=StructureType.UNKNOWN)

        def structure_to_tick_fn(structure):
            return structure.get_tick()

        index = lower_equal(
            self.get_structures(),
            tick_to_structure_fn(tick),
            key=lambda s: structure_to_tick_fn(s),
        )

        if index < 0:
            index = 0

        if index >= len(self._proto.structures):
            index = len(self._proto.structures) - 1

        if index == -1:
            return None

        return StructureMarker(song=self, proto=self._proto.structures[index])

    def create_structure(self, tick: int, type: StructureType):
        structure = StructureMarker(song=self, tick=tick, type=type)
        self._proto.structures.append(structure._proto)
        structure._proto = self._proto.structures[-1]
        if len(self._proto.structures) == 1:
            # If there is only 1 structure, move it to the start.
            structure.set_tick(0)
        self._proto.structures.sort(key=lambda x: x.tick)

    def move_structure(self, structure_index: int, move_to_tick: int):
        structure = self.get_structure_at_index(structure_index)
        if not structure:
            return
        if structure_index <= 0:
            return
        prev_structure = self.get_structure_at_index(structure_index - 1)
        if prev_structure.get_tick() == move_to_tick:
            # Moved to another structure, delete it.
            self.remove_structure(structure_index - 1)
        elif structure_index < len(self.get_structures()) - 1:
            next_structure = self.get_structure_at_index(structure_index + 1)
            if next_structure and next_structure.get_tick() == move_to_tick:
                # Moved to another time signature, delete it.
                self.remove_structure(structure_index + 1)
        structure.set_tick(move_to_tick)
        self._proto.structures.sort(key=lambda x: x.tick)

    def update_structure_at_tick(self, tick: int, type: StructureType):
        existing_structure = self.get_structure_at_tick(tick)
        if existing_structure:
            existing_structure.set_type(type)
        else:
            self.create_structure(tick, type)

    def remove_structure(self, index: int):
        if index < 0 or index >= len(self._proto.structures):
            return
        self._proto.structures.pop(index)
        if len(self._proto.structures) > 0 and self._proto.structures[0].tick > 0:
            # If the first structure of the remaining ones does not start
            # from 0, move it to 0.
            self._proto.structures[0].tick = 0
        self._proto.structures.sort(key=lambda x: x.tick)

    def serialize(self):
        return b64encode(self._proto.SerializeToString()).decode('ascii')

    def serialize_to_bytestring(self) -> str:
        '''
        Note here the returned string is essentially bytes, just using the str form for convenience.
        See https://protobuf.dev/getting-started/pythontutorial/#parsing-serialization
        '''
        return self._proto.SerializeToString()

    @staticmethod
    def deserialize(serialized_song_string: str):
        song_proto = song_pb2.Song()
        song_proto.ParseFromString(
            b64decode(serialized_song_string))  # type: ignore
        return Song(proto=song_proto)

    @staticmethod
    def deserialize_from_bytestring(serialized_song_bytestring: str):
        '''
        Note here the input string is essentially bytes, generated by `serialize_to_bytestring` or equivalent method in other languages.
        '''
        song_proto = song_pb2.Song()
        song_proto.ParseFromString(serialized_song_bytestring)
        return Song(proto=song_proto)

    @staticmethod
    def from_midi(midi_obj: MidiFile):
        '''
        TODO: Replace proto operations with builtin methods.
        '''

        def scale_int_by(value, scale_factor):
            return round(value * scale_factor)

        song = Song()
        song_proto = song._proto
        ppq_scale_factor = float(song_proto.PPQ) / \
                           float(midi_obj.ticks_per_beat)
        # Add tempos and time signatures
        song.overwrite_tempo_changes([TempoEvent(ticks=scale_int_by(
            tempo_change.time, ppq_scale_factor), bpm=tempo_change.tempo) for tempo_change in midi_obj.tempo_changes])
        song.overwrite_time_signature_changes([
            TimeSignatureEvent(
                ticks=scale_int_by(time_signature_change.time, ppq_scale_factor),
                numerator=time_signature_change.numerator,
                denominator=time_signature_change.denominator)
            for time_signature_change in midi_obj.time_signature_changes])

        # Add tracks and notes.
        song_last_tick = 0
        for index, instrument in enumerate(midi_obj.instruments):
            song_track_proto = song_proto.tracks.add(
                uuid=Track._generate_track_id(), rank=index, type=TrackType.MIDI_TRACK)
            song_track_proto.instrument.program = instrument.program
            song_track_proto.instrument.is_drum = instrument.is_drum
            track_clip_proto = song_track_proto.clips.add(
                id=Clip._generate_clip_id(), type=ClipType.MIDI_CLIP, clip_start_tick=0)
            # Add notes.
            for note in instrument.notes:
                start_tick = scale_int_by(
                    note.start, ppq_scale_factor)
                end_tick = scale_int_by(note.end, ppq_scale_factor)
                track_clip_proto.notes.add(
                    pitch=note.pitch, velocity=note.velocity, start_tick=start_tick, start_time=song.tick_to_seconds(
                        start_tick),
                    end_tick=end_tick, end_time=song.tick_to_seconds(end_tick))
            track_clip_proto.clip_start_tick = min(
                track_clip_proto.notes, key=lambda x: x.start_tick).start_tick
            track_clip_proto.clip_end_tick = max(
                track_clip_proto.notes, key=lambda x: x.end_tick).end_tick
            song_last_tick = max(
                song_last_tick, track_clip_proto.clip_end_tick)
            # Add automation.
            volume_ccs = []
            pan_ccs = []
            for control_change in instrument.control_changes:
                if control_change.number == 7:
                    # Volume CC
                    volume_ccs.append(control_change)
                elif control_change.number == 10:
                    # Pan CC
                    pan_ccs.append(control_change)
            if len(volume_ccs) == 1:
                song_track_proto.volume = volume_ccs[0].value / 127.0
            elif len(volume_ccs) > 1:
                volume_target = AutomationTarget(AutomationTargetType.VOLUME)
                volume_target_id = volume_target.to_tf_automation_target_id()
                song_track_proto.automation.targets.append(
                    volume_target._proto)
                volume_target._proto = song_track_proto.automation.targets[-1]
                volume_target_value = song_pb2.AutomationValue()
                song_track_proto.automation.target_values[volume_target_id] = volume_target_value
                for index, cc in enumerate(sorted(volume_ccs, key=lambda x: x.time)):
                    volume_target_value.points.add(tick=scale_int_by(
                        cc.time, ppq_scale_factor), value=cc.value / 127.0, id=index + 1)
            else:
                # Volume data missing from midi, set it to default.
                song_track_proto.volume = db_to_volume_value(0.0)

            if len(pan_ccs) == 1:
                song_track_proto.pan = pan_ccs[0].value - 64
            elif len(pan_ccs) > 1:
                pan_target = AutomationTarget(AutomationTargetType.PAN)
                pan_target_id = pan_target.to_tf_automation_target_id()
                song_track_proto.automation.targets.append(pan_target._proto)
                pan_target._proto = song_track_proto.automation.targets[-1]
                pan_target_value = song_pb2.AutomationValue()
                song_track_proto.automation.target_values[pan_target_id] = pan_target_value
                for index, cc in enumerate(sorted(pan_ccs, key=lambda x: x.time)):
                    pan_target_value.points.add(tick=scale_int_by(
                        cc.time, ppq_scale_factor), value=cc.value / 127.0, id=index + 1)

        song.last_tick = song_last_tick
        song.duration = song.tick_to_seconds(song_last_tick)
        return song

    def to_midi(self):
        '''
        TODO: Replace proto operations with builtin methods.
        '''
        midi_obj = MidiFile()
        midi_obj.ticks_per_beat = self.get_resolution()
        for tempo_proto in self._proto.tempos:
            midi_obj.tempo_changes.append(ToolkitTempoChange(
                tempo=tempo_proto.bpm, time=tempo_proto.ticks))
        for time_signature_proto in self._proto.time_signatures:
            midi_obj.time_signature_changes.append(
                ToolkitTimeSignature(
                    numerator=time_signature_proto.numerator,
                    denominator=time_signature_proto.denominator,
                    time=time_signature_proto.ticks))
        for track_proto in self._proto.tracks:
            if track_proto.type != TrackType.MIDI_TRACK or len(track_proto.clips) == 0:
                continue
            instrument = Instrument(program=track_proto.instrument.program,
                                    is_drum=track_proto.instrument.is_drum, name=f'Track {track_proto.rank}')
            midi_obj.instruments.append(instrument)
            # Export clips
            for clip_proto in track_proto.clips:
                if clip_proto.type != ClipType.MIDI_CLIP:
                    continue
                for note_proto in clip_proto.notes:
                    if not Clip.is_note_in_clip(
                            note_start_tick=note_proto.start_tick, note_end_tick=note_proto.end_tick,
                            clip_start_tick=clip_proto.clip_start_tick, clip_end_tick=clip_proto.clip_end_tick):
                        continue
                    instrument.notes.append(
                        ToolkitNote(
                            pitch=note_proto.pitch, velocity=note_proto.velocity,
                            start=note_proto.start_tick, end=note_proto.end_tick))
            # TODO: Export automation
        midi_obj.max_tick = self.get_last_tick()
        return midi_obj

    def get_resolution(self):
        return self._proto.PPQ

    def get_tempo_event_count(self):
        return len(self._proto.tempos)

    def get_tempo_event_at(self, index: int):
        if index < 0 or index >= len(self._proto.tempos):
            return None
        return TempoEvent(proto=self._proto.tempos[index])

    def get_tempo_event_at_tick(self, tick: int):
        target_tempo = SimpleNamespace()
        target_tempo.ticks = tick
        index = lower_equal(
            self._proto.tempos,
            target_tempo,
            lambda x: x.ticks,
        )
        if index < 0:
            index = 0

        if index >= len(self._proto.tempos):
            index = len(self._proto.tempos) - 1

        return TempoEvent(proto=self._proto.tempos[index])

    def create_tempo_change(self, ticks: int, bpm: float):
        '''
        Adds a tempo change event into the song and returns it.

        @param ticks The tick at which this event happens.
        @param bpm The new tempo in BPM(Beats-per-minute) format.
        '''
        if self.get_resolution() <= 0:
            raise Exception(
                'Song resolution must be provided before creating tempo changes.')

        if self.get_tempo_event_count() == 0 and ticks != 0:
            raise Exception('The first tempo event must be at tick 0')

        # Calculate time BEFORE the new tempo event is inserted.
        tempo_change = TempoEvent(
            ticks=ticks, bpm=bpm, time=self.tick_to_seconds(ticks))
        insert_index = greater_equal(
            self._proto.tempos,
            tempo_change._proto,
            lambda x: x.ticks
        )
        if insert_index < 0:
            self._proto.tempos.append(tempo_change._proto)
            tempo_change._proto = self._proto.tempos[-1]
        else:
            self._proto.tempos.insert(insert_index, tempo_change._proto)
            tempo_change._proto = self._proto.tempos[insert_index]

        self.retiming_tempo_events()
        return tempo_change

    def move_tempo(self, tempo_index: int, move_to_tick: int):
        tempo = self.get_tempo_event_at(tempo_index)
        if tempo is None:
            return

        if (tempo_index == 0):
            # Cannot move the first tempo.
            return

        prev_tempo = self.get_tempo_event_at(tempo_index - 1)
        if prev_tempo is None:
            return
        if (prev_tempo.get_ticks() == move_to_tick):
            # Moved to another tempo, delete it.
            self.remove_tempo_change_at(tempo_index - 1)
        elif (tempo_index < self.get_tempo_event_count() - 1):
            next_tempo = self.get_tempo_event_at(tempo_index + 1)
            if (next_tempo is not None and next_tempo.get_ticks() == move_to_tick):
                # Moved to another tempo, delete it.
                self.remove_tempo_change_at(tempo_index + 1)

        tempo.set_ticks(move_to_tick)
        self.retiming_tempo_events()

    def remove_tempo_change_at(self, index: int):
        if self.get_tempo_event_count() <= 1:
            raise Exception('Song has to have at least one tempo change. Update the last tempo change instead.')

        if (index == 0):
            raise Exception('Cannot remove the first tempo.')

        self._proto.tempos.pop(index)
        self.retiming_tempo_events()

    def retiming_tempo_events(self):
        sorted_tempos = sorted(
            self._proto.tempos, key=lambda tempo: tempo.ticks)
        del self._proto.tempos[:]
        self._proto.tempos.extend(sorted_tempos)
        # Re-calculate all tempo event time.
        for tempo_event_proto in self._proto.tempos:
            tempo_event_proto.time = self.tick_to_seconds(
                tempo_event_proto.ticks)

    def tick_to_seconds(self, tick: int):
        if tick == 0:
            return 0

        target_tempo = SimpleNamespace()
        target_tempo.ticks = tick
        base_tempo_index = lower_than(
            self._proto.tempos,
            target_tempo,
            lambda x: x.ticks
        )
        if base_tempo_index == -1:
            # If no tempo is found before the tick, use the first tempo.
            base_tempo_index = 0

        base_tempo_change = self._proto.tempos[base_tempo_index]
        ticks_delta = tick - base_tempo_change.ticks
        ticks_per_second_since_last_tempo_change = Song._tempo_bpm_to_ticks_per_second(
            base_tempo_change.bpm,
            self.get_resolution(),
        )
        return base_tempo_change.time + ticks_delta / ticks_per_second_since_last_tempo_change

    def seconds_to_tick(self, seconds: float):
        if (seconds == 0):
            return 0

        target_tempo = SimpleNamespace()
        target_tempo.time = seconds
        base_tempo_index = lower_than(
            self._proto.tempos,
            target_tempo,
            key=lambda x: x.time
        )
        if (base_tempo_index == -1):
            # If no tempo is found before the time, use the first tempo.
            base_tempo_index = 0

        base_tempo_change_proto = self._proto.tempos[base_tempo_index]
        time_delta = seconds - base_tempo_change_proto.time
        ticks_per_second_since_last_tempo_change = Song._tempo_bpm_to_ticks_per_second(
            base_tempo_change_proto.bpm,
            self.get_resolution(),
        )
        return round(base_tempo_change_proto.ticks + time_delta * ticks_per_second_since_last_tempo_change)

    def overwrite_tempo_changes(self, tempo_events: List[TempoEvent]):
        if len(tempo_events) == 0:
            raise Exception('Cannot clear all the tempo events.')
        sorted_tempo_events = sorted(
            tempo_events, key=lambda tempo_event: tempo_event.get_ticks())
        first_tempo_event = sorted_tempo_events[0]
        if first_tempo_event.get_ticks() > 0:
            raise Exception('The first tempo event needs to start from tick 0')
        del self._proto.tempos[:]
        self._proto.tempos.add(
            ticks=0, time=0, bpm=first_tempo_event.get_bpm())
        for i in range(1, len(sorted_tempo_events)):
            tempo_event = sorted_tempo_events[i]
            self.create_tempo_change(
                ticks=tempo_event.get_ticks(), bpm=tempo_event.get_bpm())
        self.retiming_tempo_events()

    def overwrite_time_signature_changes(self, time_signatures: List[TimeSignatureEvent]):
        if len(time_signatures) == 0:
            raise Exception('At least one time signature needs to be present.')
        del self._proto.time_signatures[:]
        for time_signature_change in time_signatures:
            self._proto.time_signatures.add(
                ticks=time_signature_change.get_ticks(),
                numerator=time_signature_change.get_numerator(),
                denominator=time_signature_change.get_denominator())

    def get_time_signature_event_count(self):
        return len(self._proto.time_signatures)

    def get_time_signature_event_at(self, index: int):
        return TimeSignatureEvent(proto=self._proto.time_signatures[index])

    def get_time_signature_event_at_tick(self, tick: int):
        target_time_signature = SimpleNamespace()
        target_time_signature.ticks = tick
        index = lower_equal(
            self._proto.time_signatures,
            target_time_signature,
            lambda x: x.ticks,
        )
        if index < 0:
            index = 0

        if index >= len(self._proto.time_signatures):
            index = len(self._proto.time_signatures) - 1

        return TimeSignatureEvent(proto=self._proto.time_signatures[index])

    def create_time_signature(self, ticks: int, numerator: int, denominator: int):
        '''
        @param ticks The tick at which this event happens.
        '''
        time_signature_proto = TimeSignatureEvent(ticks=ticks, numerator=numerator, denominator=denominator)._proto
        target_time_signature = SimpleNamespace()
        target_time_signature.ticks = ticks
        insert_index = greater_equal(
            self._proto.time_signatures,
            target_time_signature,
            lambda x: x.ticks
        )
        if (insert_index < 0):
            self._proto.time_signatures.append(time_signature_proto)
            return TimeSignatureEvent(proto=self._proto.time_signatures[-1])
        else:
            self._proto.time_signatures.insert(insert_index, time_signature_proto)
            return TimeSignatureEvent(proto=self._proto.time_signatures[insert_index])

    def create_audio_plugin(self, tf_id: str):
        pluginInfo = decode_audio_plugin_tuneflow_id(tf_id)
        plugin = AudioPlugin(
            name=pluginInfo["name"],
            manufacturer_name=pluginInfo["manufacturer_name"],
            plugin_format_name=pluginInfo["plugin_format_name"],
            plugin_version=pluginInfo["plugin_version"],
        )
        return plugin

    def create_track(self, type: int,
                     index: int | None = None,
                     rank: int | None = None,
                     assign_default_sampler_plugin=False):
        '''
        Adds a new track into the song and returns it.

        @param index Index to insert at. If left blank, appends to the end.
        @param rank The displayed rank which uniquely identifies a track. Internal use, do not set this.
        @param assign_default_sampler_plugin Whether to assign a default sampler plugin if type is `MIDI_TRACK`.
        '''
        new_track = Track(
            type=type, song=self, rank=rank if rank is not None else self.get_next_track_rank())
        if assign_default_sampler_plugin and type == TrackType.MIDI_TRACK:
            new_track.set_sampler_plugin(new_track.create_audio_plugin(AudioPlugin.DEFAULT_SYNTH_TFID))
        if type == TrackType.AUX_TRACK:
            # TODO: Set default input bus rank.
            pass
            # new_track.getAuxTrackData().setInputBusRank(1)

        if index is None:
            index = len(self._proto.tracks)
        self._proto.tracks.insert(index, new_track._proto)
        new_track._proto = self._proto.tracks[index]
        return new_track

    def get_next_track_rank(self):
        return 1 if len(self._proto.tracks) == 0 else max([track.rank for track in self._proto.tracks]) + 1

    def clone_track(self, track: Track) -> Track:
        '''
        Clones a track and inserts it in this song and returns the cloned instance.
        '''
        new_proto = song_pb2.Track()
        new_proto.CopyFrom(track._proto)
        new_proto.rank = self.get_next_track_rank()
        new_proto.uuid = Track._generate_track_id()
        self._proto.tracks.insert(self.get_track_index(track.get_id()), new_proto)
        return self.get_track_by_id(new_proto.uuid)

    def __repr__(self) -> str:
        return str(self._proto)

    @staticmethod
    def _tempo_bpm_to_ticks_per_second(tempo_bpm: float, PPQ: int):
        return (tempo_bpm * PPQ) / 60

    @staticmethod
    def get_default_resolution():
        '''
        Returns the default Pulse-per-Quater-Note used in TuneFlow.
        '''
        return 480



from tuneflow_py import Song, TrackType, TrackOutputType
from miditoolkit.midi import MidiFile
from pathlib import PurePath, Path
import unittest
import pytest


def create_song():
    song = Song()
    song.create_tempo_change(
        ticks=1440,
        bpm=60,
    )
    song.create_time_signature(ticks=2880, numerator=7, denominator=8)
    return song


class BaseTest(unittest.TestCase):
    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)

        self.song = create_song()


class TestImportExportMIDI(unittest.TestCase):
    def test_import_midi(self):
        midi_obj = MidiFile(filename=PurePath(
            Path(__file__).parent, Path('caravan.test.golden.mid')))
        song = Song.from_midi(midi_obj=midi_obj)
        # Test tempos
        self.assertEqual(song.get_tempo_event_count(), 8)
        self.assertAlmostEqual(song.get_duration(), 595.0945734687816)
        # Test time signatures
        self.assertEqual(song.get_time_signature_event_count(), 3)
        self.assertEqual(
            song.get_time_signature_event_at(1).get_numerator(), 4)
        self.assertEqual(song.get_time_signature_event_at(
            1).get_denominator(), 2)
        # Test tracks
        first_track = song.get_track_at(0)
        self.assertEqual(first_track.get_clip_count(), 1)
        # Test non-piano track
        # Test track volume, pan
        self.assertAlmostEqual(
            first_track.get_volume_in_db(), -0.6665511534374842)
        self.assertAlmostEqual(song.get_track_at(7).get_pan(), 63)
        # Test track instrument
        self.assertEqual(first_track.get_instrument().program, 65)  # type:ignore
        self.assertEqual(first_track.get_instrument().is_drum, False)  # type:ignore
        # Test drum track
        last_track = song.get_track_at(song.get_track_count()-1)
        self.assertEqual(last_track.get_instrument().program, 0)  # type:ignore
        self.assertEqual(last_track.get_instrument().is_drum, True)  # type:ignore
        # TODO: Test track automation
        first_clip = first_track.get_clip_at(0)
        self.assertEqual(first_clip.get_raw_note_count(), 673)
        self.assertEqual(first_clip.get_clip_start_tick(), 107520)
        self.assertEqual(first_clip.get_clip_end_tick(), 1326719)
        first_note = first_clip.get_raw_note_at(0)
        self.assertEqual(first_note.get_pitch(), 52)
        self.assertEqual(first_note.get_start_tick(), 107520)
        self.assertEqual(first_note.get_end_tick(), 108159)
        self.assertEqual(first_note.get_velocity(), 115)
        self.assertEqual(first_note.get_start_time(), 47.659584045410156)
        self.assertAlmostEqual(first_note.get_end_time(), 47.94282913208008)
        self.assertEqual(song.last_tick, 1327199)
        self.assertAlmostEqual(song.duration, 595.0945734687816)

    def test_export_midi(self):
        golden_midi_path = PurePath(
            Path(__file__).parent, Path('caravan.test.golden.mid'))
        export_midi_path = PurePath(
            Path(__file__).parent, Path('caravan.test.test.mid'))
        midi_obj = MidiFile(filename=golden_midi_path)
        song = Song.from_midi(midi_obj=midi_obj)
        export_file = open(export_midi_path, 'wb')
        exported_midi = song.to_midi()
        exported_midi.dump(file=export_file)
        export_file.close()

        self.assertEqual(len(midi_obj.tempo_changes),
                         len(exported_midi.tempo_changes))
        for i in range(len(midi_obj.tempo_changes)):
            expected_tempo = midi_obj.tempo_changes[i]
            actual_tempo = exported_midi.tempo_changes[i]
            self.assertEqual(expected_tempo.time, actual_tempo.time)
            self.assertAlmostEqual(expected_tempo.tempo,
                                   actual_tempo.tempo, places=4)
        self.assertEqual(len(midi_obj.time_signature_changes),
                         len(exported_midi.time_signature_changes))
        for i in range(len(midi_obj.time_signature_changes)):
            expected_time_signature = midi_obj.time_signature_changes[i]
            actual_time_signature = exported_midi.time_signature_changes[i]
            self.assertEqual(expected_time_signature.time,
                             actual_time_signature.time)
            self.assertEqual(expected_time_signature.numerator,
                             actual_time_signature.numerator)
            self.assertEqual(expected_time_signature.denominator,
                             actual_time_signature.denominator)

        self.assertEqual(len(midi_obj.instruments),
                         len(exported_midi.instruments))
        for i in range(len(midi_obj.instruments)):
            expected_track = midi_obj.instruments[i]
            actual_track = exported_midi.instruments[i]
            self.assertEqual(expected_track.program, actual_track.program)
            self.assertEqual(expected_track.is_drum, actual_track.is_drum)
            self.assertEqual(len(expected_track.notes),
                             len(actual_track.notes))
            for j in range(len(expected_track.notes)):
                expected_note = expected_track.notes[j]
                actual_note = actual_track.notes[j]
                self.assertEqual(expected_note.start, actual_note.start)
                self.assertEqual(expected_note.end, actual_note.end)
                self.assertEqual(expected_note.velocity, actual_note.velocity)
                self.assertEqual(expected_note.pitch, actual_note.pitch)
            # TODO: Test track automation.


class TestBasicOperations(BaseTest):
    def test_get_track_index(self):
        track = self.song.create_track(type=TrackType.MIDI_TRACK)
        track2 = self.song.create_track(type=TrackType.MIDI_TRACK)
        self.assertEqual(self.song.get_track_index(track.get_id()), 0)
        self.assertEqual(self.song.get_track_index(track2.get_id()), 1)


class TestTempo(BaseTest):
    def test_get_tempo_at(self):
        song = self.song
        self.assertEqual(song.get_tempo_event_at(0).get_ticks(), 0)  # type:ignore
        self.assertEqual(song.get_tempo_event_at(0).get_bpm(), 120)  # type:ignore
        self.assertEqual(song.get_tempo_event_at(1).get_ticks(), 1440)  # type:ignore
        self.assertEqual(song.get_tempo_event_at(1).get_bpm(), 60)  # type:ignore
        self.assertIsNone(song.get_tempo_event_at(-1))
        self.assertIsNone(song.get_tempo_event_at(2))

    def test_get_tempo_at_tick(self):
        song = self.song
        self.assertEqual(song.get_tempo_event_at_tick(-10).get_ticks(), 0)
        self.assertEqual(song.get_tempo_event_at_tick(-10).get_bpm(), 120)
        self.assertEqual(song.get_tempo_event_at_tick(0).get_ticks(), 0)
        self.assertEqual(song.get_tempo_event_at_tick(0).get_bpm(), 120)
        self.assertEqual(song.get_tempo_event_at_tick(10).get_ticks(), 0)
        self.assertEqual(song.get_tempo_event_at_tick(10).get_bpm(), 120)
        self.assertEqual(song.get_tempo_event_at_tick(1439).get_ticks(), 0)
        self.assertEqual(song.get_tempo_event_at_tick(1439).get_bpm(), 120)
        self.assertEqual(song.get_tempo_event_at_tick(1440).get_ticks(), 1440)
        self.assertEqual(song.get_tempo_event_at_tick(1440).get_bpm(), 60)
        self.assertEqual(song.get_tempo_event_at_tick(1441).get_ticks(), 1440)
        self.assertEqual(song.get_tempo_event_at_tick(1441).get_bpm(), 60)
        self.assertEqual(song.get_tempo_event_at_tick(9999).get_ticks(), 1440)
        self.assertEqual(song.get_tempo_event_at_tick(9999).get_bpm(), 60)

    def test_move_tempo_non_overlapping(self):
        song = self.song
        song.create_tempo_change(
            ticks=2880,
            bpm=240,
        )
        self.assertEqual(song.get_tempo_event_count(), 3)
        with pytest.raises(Exception) as exception:
            song.move_tempo(1, 0)
        self.assertIsNotNone(exception)
        song.move_tempo(1, 480)
        self.assertEqual(song.get_tempo_event_count(), 3)
        self.assertEqual(song.get_tempo_event_at(0).get_bpm(), 120)  # type:ignore
        self.assertEqual(song.get_tempo_event_at(0).get_time(), 0)  # type:ignore
        self.assertEqual(song.get_tempo_event_at(1).get_bpm(), 60)  # type:ignore
        self.assertAlmostEqual(song.get_tempo_event_at(1).get_time(), 0.5)  # type:ignore
        self.assertEqual(song.get_tempo_event_at(2).get_bpm(), 240)  # type:ignore
        self.assertAlmostEqual(song.get_tempo_event_at(2).get_time(), 5.5)  # type:ignore

    def test_move_tempo_past_another(self):
        song = self.song
        song.create_tempo_change(
            ticks=2880,
            bpm=240,
        )
        self.assertEqual(song.get_tempo_event_count(), 3)
        song.move_tempo(1, 3360)
        self.assertEqual(song.get_tempo_event_count(), 3)
        self.assertEqual(song.get_tempo_event_at(0).get_bpm(), 120)  # type:ignore
        self.assertEqual(song.get_tempo_event_at(0).get_time(), 0)  # type:ignore
        self.assertEqual(song.get_tempo_event_at(1).get_bpm(), 240)  # type:ignore
        self.assertAlmostEqual(song.get_tempo_event_at(1).get_time(), 3)  # type:ignore
        self.assertEqual(song.get_tempo_event_at(2).get_bpm(), 60)  # type:ignore
        self.assertAlmostEqual(song.get_tempo_event_at(2).get_time(), 3.25)  # type:ignore

    def test_move_tempo_overwrite(self):
        song = self.song
        song.create_tempo_change(
            ticks=2880,
            bpm=240,
        )
        self.assertEqual(song.get_tempo_event_count(), 3)
        with pytest.raises(Exception) as exception:
            song.move_tempo(1, 0)
        self.assertIsNotNone(exception)
        song.move_tempo(2, 1440)
        self.assertEqual(song.get_tempo_event_count(), 2)
        self.assertEqual(song.get_tempo_event_at(0).get_bpm(), 120)  # type:ignore
        self.assertEqual(song.get_tempo_event_at(1).get_bpm(), 240)  # type:ignore

    def test_remove_tempo(self):
        song = self.song
        with pytest.raises(Exception) as exception:
            song.remove_tempo_change_at(0)
        self.assertIsNotNone(exception)
        song.remove_tempo_change_at(1)
        self.assertEqual(song.get_tempo_event_count(), 1)
        self.assertEqual(song.get_tempo_event_at(0).get_bpm(), 120)  # type:ignore
        with pytest.raises(Exception) as exception2:
            song.remove_tempo_change_at(0)
        self.assertIsNotNone(exception2)


class TestTimeSignature(BaseTest):
    def test_get_time_signature(self):
        self.assertEqual(self.song.get_time_signature_event_count(), 2)
        time_signature1 = self.song.get_time_signature_event_at(0)
        self.assertEqual(time_signature1.get_ticks(), 0)
        self.assertEqual(time_signature1.get_numerator(), 4)
        self.assertEqual(time_signature1.get_denominator(), 4)
        time_signature2 = self.song.get_time_signature_event_at(1)
        self.assertEqual(time_signature2.get_ticks(), 2880)
        self.assertEqual(time_signature2.get_numerator(), 7)
        self.assertEqual(time_signature2.get_denominator(), 8)
        time_signature = self.song.get_time_signature_event_at_tick(-1)
        self.assertEqual(time_signature.get_ticks(), 0)
        self.assertEqual(time_signature.get_numerator(), 4)
        self.assertEqual(time_signature.get_denominator(), 4)
        time_signature = self.song.get_time_signature_event_at_tick(0)
        self.assertEqual(time_signature.get_ticks(), 0)
        self.assertEqual(time_signature.get_numerator(), 4)
        self.assertEqual(time_signature.get_denominator(), 4)
        time_signature = self.song.get_time_signature_event_at_tick(1440)
        self.assertEqual(time_signature.get_ticks(), 0)
        self.assertEqual(time_signature.get_numerator(), 4)
        self.assertEqual(time_signature.get_denominator(), 4)
        time_signature = self.song.get_time_signature_event_at_tick(2880)
        self.assertEqual(time_signature.get_ticks(), 2880)
        self.assertEqual(time_signature.get_numerator(), 7)
        self.assertEqual(time_signature.get_denominator(), 8)
        time_signature = self.song.get_time_signature_event_at_tick(99999)
        self.assertEqual(time_signature.get_ticks(), 2880)
        self.assertEqual(time_signature.get_numerator(), 7)
        self.assertEqual(time_signature.get_denominator(), 8)


class TestTracks(BaseTest):
    def test_clone_track(self):
        track1 = self.song.create_track(type=TrackType.MIDI_TRACK, index=0, rank=1)
        self.assertEqual(self.song.get_track_count(), 1)
        track1.set_instrument(program=64, is_drum=False)
        track1.set_pan(62)
        track2 = self.song.clone_track(track1)
        self.assertEqual(self.song.get_track_count(), 2)
        self.assertNotEqual(track1.get_id(), track2.get_id())
        self.assertEqual(self.song.get_track_index(track1.get_id()), 1)
        self.assertEqual(self.song.get_track_index(track2.get_id()), 0)
        self.assertEqual(track1.get_rank(), 1)
        self.assertEqual(track2.get_rank(), 2)
        self.assertEqual(track2.get_instrument().program, 64)  # type:ignore
        self.assertEqual(track2.get_instrument().is_drum, False)  # type:ignore
        self.assertEqual(track2.get_pan(), 62)

    def test_remove_track(self):
        track1 = self.song.create_track(type=TrackType.MIDI_TRACK)
        dep_track = self.song.create_track(type=TrackType.AUX_TRACK)
        dep_track.get_or_create_output().set_type(TrackOutputType.TRACK_OUTPUT_TRACK)
        dep_track.get_or_create_output().set_track_id(track1.get_id())
        self.assertEqual(self.song.get_track_count(), 2)
        self.assertTrue(dep_track.has_output())
        self.assertEqual(dep_track.get_output().get_track_id(), track1.get_id())  # type:ignore
        self.assertEqual(dep_track.get_output().get_type(), TrackOutputType.TRACK_OUTPUT_TRACK)  # type:ignore
        self.song.remove_track(track1.get_id())
        self.assertEqual(self.song.get_track_count(), 1)
        self.assertFalse(dep_track.has_output())


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from tuneflow_py.base_plugin import TuneflowPlugin
from tuneflow_py.models.audio_plugin import AudioPlugin, get_audio_plugin_tuneflow_id, are_tuneflow_ids_equal, are_tuneflow_ids_equal_ignore_version, decode_audio_plugin_tuneflow_id
from tuneflow_py.models.automation import AutomationTarget, AutomationTargetType, AutomationData, AutomationPoint, AutomationValue
from tuneflow_py.models.clip import ClipType, Clip
from tuneflow_py.models.note import Note
from tuneflow_py.models.song import Song
from tuneflow_py.models.tempo import TempoEvent
from tuneflow_py.models.marker import StructureMarker, StructureType
from tuneflow_py.models.time_signature import TimeSignatureEvent
from tuneflow_py.models.track import TrackType, Track, TrackOutputType
from tuneflow_py.descriptors.widget import *
from tuneflow_py.descriptors.text import *
from tuneflow_py.descriptors.param import *
from tuneflow_py.descriptors.common import *
from tuneflow_py.descriptors.audio_plugin_descriptor import *
from tuneflow_py.descriptors.clip_descriptor import *
from tuneflow_py.descriptors.plugin import *
from tuneflow_py.utils import *



from tuneflow_py.models.song import Song, TrackType
from tuneflow_py.models.audio_plugin import AudioPlugin
from tuneflow_py.utils import db_to_volume_value
import unittest


def create_song():
    song = Song()
    song.create_tempo_change(ticks=1440, bpm=60)
    return song


class TestCreateTracks(unittest.TestCase):
    def test_create_midi_track(self):
        song = create_song()
        track = song.create_track(type=TrackType.MIDI_TRACK)
        self.assertAlmostEqual(track.get_volume(), db_to_volume_value(0))
        self.assertEqual(track.get_pan(), 0)
        self.assertEqual(track.get_solo(), False)
        self.assertEqual(track.get_muted(), False)
        self.assertEqual(track.get_type(), TrackType.MIDI_TRACK)

    def test_create_audio_track(self):
        song = create_song()
        track = song.create_track(type=TrackType.AUDIO_TRACK)
        self.assertAlmostEqual(track.get_volume(), db_to_volume_value(0))
        self.assertEqual(track.get_pan(), 0)
        self.assertEqual(track.get_solo(), False)
        self.assertEqual(track.get_muted(), False)
        self.assertEqual(track.get_type(), TrackType.AUDIO_TRACK)

    def test_create_track_returns_correct_reference(self):
        song = create_song()
        self.assertEqual(song.get_track_count(), 0)
        new_track = song.create_track(type=TrackType.MIDI_TRACK)
        self.assertEqual(new_track.get_instrument().program, 0)  # type:ignore
        self.assertEqual(song.get_track_at(0).get_instrument().program, 0)  # type:ignore
        new_track.get_instrument().program = 127  # type:ignore
        self.assertEqual(new_track.get_instrument().program, 127)  # type:ignore
        self.assertEqual(song.get_track_at(0).get_instrument().program, 127)  # type:ignore


class TestAudioTracks(unittest.TestCase):
    def test_cannot_set_instrument_or_suggested_instrument_or_sampler_plugin(self):
        song = create_song()
        track = song.create_track(type=TrackType.AUDIO_TRACK)
        self.assertFalse(track.has_instrument())
        self.assertFalse(track.has_sampler_plugin())
        self.assertEqual(track.get_suggested_instruments_count(), 0)
        track.set_instrument(program=32, is_drum=False)
        self.assertFalse(track.has_instrument())
        track.set_sampler_plugin(AudioPlugin(name="plugin1", manufacturer_name="manufacturer1",
                                 plugin_format_name="VST", plugin_version="1.1"))
        self.assertIsNone(track.get_sampler_plugin())
        self.assertFalse(track.has_sampler_plugin())
        track.create_suggested_instrument(program=64, is_drum=False)
        self.assertEqual(track.get_suggested_instruments_count(), 0)

    def test_clone_audio_track(self):
        # TODO: Test
        pass


class TestMIDITracks(unittest.TestCase):
    def test_clone_midi_track(self):
        # TODO: Test
        pass


class TestCloneClip(unittest.TestCase):
    def test_clone_clip(self):
        song = create_song()
        track = song.create_track(type=TrackType.MIDI_TRACK)
        clip1 = track.create_midi_clip(clip_start_tick=10, clip_end_tick=20, insert_clip=True)
        clip1.create_note(pitch=72, velocity=100, start_tick=15, end_tick=16)
        clip2 = track.clone_clip(clip1)
        self.assertNotEqual(clip1.get_id(), clip2.get_id())
        self.assertEqual(clip2.get_clip_start_tick(), clip1.get_clip_start_tick())
        self.assertEqual(clip2.get_clip_end_tick(), clip1.get_clip_end_tick())
        self.assertEqual(clip2.get_raw_note_at(0).get_pitch(), 72)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from tuneflow_py.descriptors.clip_descriptor import AudioClipData
from tuneflow_py.models.protos import song_pb2
from tuneflow_py.models.note import Note
from tuneflow_py.utils import lower_than, greater_than, greater_equal
from nanoid import generate as generate_nanoid
from typing import List
from types import SimpleNamespace


ClipType = song_pb2.ClipType


class Clip:
    def __init__(self, song, type: int | None = None, clip_start_tick: int | None = None, id: str | None = None, track=None,
                 clip_end_tick: int | None = None, audio_clip_data: AudioClipData | None = None, proto: song_pb2.Clip |
                 None = None) -> None:
        '''
        A clip is a piece in a track, and it contains notes and the clip range.
        One track can contain one or many non-overlapping clips.

        IMPORTANT: Do not use the constructor directly, call `createClip` from tracks instead.

        @param song The song that the clip belongs to, should always be provided.
        @param track If the clip is dettached from any track, this field is going to be None.
        @param clip_start_tick Inclusive start tick.
        @param clip_end_tick Inclusive end tick.
        @param audio_clip_data Audio related data if the clip type is AUDIO_CLIP.
        '''
        self.song = song
        self.track = track
        self._next_note_id = None
        if proto is not None:
            self._proto = proto
            return
        self._proto = song_pb2.Clip()
        self._proto.id = id if id is not None else Clip._generate_clip_id()
        self._proto.type = type
        if type == ClipType.AUDIO_CLIP:
            if audio_clip_data is None:
                raise Exception(
                    'Audio clip data must be provided for audio clips.')

            if "audio_file_path" in audio_clip_data and audio_clip_data["audio_file_path"] is not None and audio_clip_data["audio_file_path"] != "":
                self._proto.audio_clip_data.audio_file_path = audio_clip_data["audio_file_path"]
            if "audio_data" in audio_clip_data and audio_clip_data["audio_data"] is not None:
                self._proto.audio_clip_data.audio_data.format = audio_clip_data["audio_data"]["format"]
                self._proto.audio_clip_data.audio_data.data = audio_clip_data["audio_data"]["data"]
            self._proto.audio_clip_data.start_tick = audio_clip_data["start_tick"]
            self._proto.audio_clip_data.duration = audio_clip_data["duration"]

            clip_start_tick = max(
                clip_start_tick, audio_clip_data["start_tick"]) if clip_start_tick is not None else audio_clip_data["start_tick"]
            audio_end_tick = self.get_audio_end_tick()
            if (clip_end_tick is None or audio_end_tick < clip_end_tick):  # type:ignore
                clip_end_tick = audio_end_tick

            self._proto.clip_start_tick = clip_start_tick
            self._proto.clip_end_tick = clip_end_tick
        elif (type == ClipType.MIDI_CLIP):
            self._proto.clip_start_tick = clip_start_tick
            if clip_end_tick is None:
                raise Exception(
                    'clip end tick must be provided when creating MIDI clip.')

            self._proto.clip_end_tick = clip_end_tick

    def get_id(self):
        return self._proto.id

    def get_track(self):
        return self.track

    def get_audio_end_tick(self) -> int | None:
        '''
        Gets the current clip audio's end tick.

        Returns None if the clip is not audio clip or audio clip data is missing.
        '''
        if self.get_type() != ClipType.AUDIO_CLIP or not self.has_audio_clip_data():
            return None

        duration = self.get_audio_duration()
        if duration is None:
            return None

        audio_start_time = self.song.tick_to_seconds(
            self._proto.audio_clip_data.start_tick)
        return self.song.seconds_to_tick(audio_start_time + duration)

    def get_duration(self):
        return self.song.tick_to_seconds(
            self.get_clip_end_tick()) - self.song.tick_to_seconds(
            self.get_clip_start_tick())

    def get_audio_duration(self) -> float | None:
        '''
        Gets the audio's duration if the clip is AUDIO_CLIP.

        Returns None if the clip is not AUDIO_CLIP or audio clip data is missing.
        '''
        if self.get_type() != ClipType.AUDIO_CLIP or not self.has_audio_clip_data():
            return None
        return self._proto.audio_clip_data.duration

    def get_raw_note_count(self):
        return len(self._proto.notes)

    def get_raw_notes(self):
        '''
        @returns All notes contained by the clip, including those that
        are not within the clip's range.
        '''
        for i in range(len(self._proto.notes)):
            yield self.get_raw_note_at(i)

    def get_raw_note_at(self, index: int):
        return Note(proto=self._proto.notes[index], clip=self)

    def get_notes(self):
        '''
        @returns Notes within the clip's range.
        '''

        note_protos = Clip._get_notes_in_range(
            raw_notes=self._proto.notes, start_tick=self.get_clip_start_tick(),
            end_tick=self.get_clip_end_tick())
        for note_proto in note_protos:
            yield Note(proto=note_proto, clip=self)

    def create_note(
        self,
        pitch: int,
        velocity: int,
        start_tick: int,
        end_tick: int,
        update_clip_range: bool = True,
        resolve_clip_conflict: bool = True
    ):
        '''
        Adds a note to the clip and returns it.

        @param pitch An integer value between 0 - 127
        @param velocity An integer value between 0 - 127
        @param start_tick An integer value indicating the start tick.
        @param end_tick An integer value indicating the end tick.
        @param update_clip_range Whether to update the clip's range if the note stretches outside the clip.
        @param resolve_clip_conflict Whether to resolve clip conflict if the clip range is updated.
        '''
        if self.get_type() != ClipType.MIDI_CLIP:
            # Only MIDI clips can create notes.
            return None
        if (
            not Note.is_valid_pitch(pitch) or
            not Note.is_note_range_valid(start_tick, end_tick) or
            not Note.is_note_velocity_valid(velocity)
        ):
            return None

        note = Note(
            pitch=pitch,
            velocity=velocity,
            start_tick=start_tick,
            end_tick=end_tick,
            id=self._get_next_note_id(),
        )
        if (start_tick < self.get_clip_start_tick() and update_clip_range):
            self.adjust_clip_left(start_tick, resolve_clip_conflict)

        if (end_tick > self.get_clip_end_tick() and update_clip_range):
            self.adjust_clip_right(end_tick, resolve_clip_conflict)

        self._ordered_insert_note(note)
        return note

    def get_type(self) -> int:
        return self._proto.type

    def delete_note(self, note: Note):
        note_index = self._get_note_index(note)
        if (note_index >= 0):
            self.delete_note_at(note_index)
            note.clip = None

    def delete_note_at(self, index: int):
        if (index >= 0 and index < len(self._proto.notes)):
            self._proto.notes.pop(index)

    def get_clip_start_tick(self) -> int:
        return self._proto.clip_start_tick

    def get_clip_end_tick(self) -> int:
        return self._proto.clip_end_tick

    def has_audio_clip_data(self) -> bool:
        return self._proto.HasField("audio_clip_data")

    def get_audio_clip_data(self):
        if not self.has_audio_clip_data():
            return None
        return self._proto.audio_clip_data

    def clear_notes(self):
        del self._proto.notes[:]

    def delete_from_parent(self, delete_associated_track_automation: bool):
        if self.track is not None:
            self.track.delete_clip(self, delete_associated_track_automation)
            self.track = None

    def adjust_clip_left(self, clip_start_tick: int, resolve_conflict=True):
        '''
        Adjust the left boundary of the clip.

        NOTE: This could delete the clip if the range becomes empty after
        this call.

        @param clip_start_tick The new start tick (inclusive) of the clip.
        '''
        clip_start_tick = max(0, clip_start_tick)
        if self.get_type() == ClipType.AUDIO_CLIP and self.has_audio_clip_data():
            # Clip boundary cannot exceed the audio's boundary.
            clip_start_tick = max(
                clip_start_tick, self._proto.audio_clip_data.start_tick)

        if clip_start_tick > self.get_clip_end_tick():
            self.delete_from_parent(delete_associated_track_automation=True)
        else:
            # Resolve conflict before changing the clip's range
            # to preserve the current order of clips.
            if (resolve_conflict and self.track is not None):
                self.track._resolve_clip_conflict(
                    self.get_id(),
                    min(self.get_clip_start_tick(), clip_start_tick),
                    self.get_clip_end_tick(),
                )
            self._proto.clip_start_tick = clip_start_tick

    def adjust_clip_right(self, clip_end_tick: int, resolve_conflict=True):
        '''
        Adjust the right boundary of the clip.

        NOTE: This could delete the clip if the range becomes empty after
        this call.

        @param clipEndTick  The new end tick (inclusive) of the clip.
        '''
        if self.get_type() == ClipType.AUDIO_CLIP:
            # Clip right boundary cannot exceed the audio's right boundary.
            audio_end_tick = self.get_audio_end_tick()
            if audio_end_tick is not None:
                clip_end_tick = min(clip_end_tick, audio_end_tick)
        if clip_end_tick < self.get_clip_start_tick() or clip_end_tick < 0:
            self.delete_from_parent(delete_associated_track_automation=True)
        else:
            # Resolve conflict before changing the clip's range
            # to preserve the current order of clips.
            if (resolve_conflict and self.track is not None):
                self.track._resolve_clip_conflict(
                    self.get_id(),
                    self.get_clip_start_tick(),
                    max(self.get_clip_end_tick(), clip_end_tick),
                )

            self._proto.clip_end_tick = clip_end_tick

    def move_clip(self, offset_tick: int, move_associated_track_automation_points: bool):
        '''
        Move the clip by the given offset ticks.

        NOTE: Moving a clip can break the order of clips,
        so the clip must be re-positioned to the correct index.
        '''
        new_clip_start_tick = max(0, self.get_clip_start_tick() + offset_tick)
        new_clip_end_tick = self.get_clip_end_tick() + offset_tick
        if (new_clip_end_tick < 0):
            self.delete_from_parent(delete_associated_track_automation=True)
            return
        if self.track is not None:
            # Resolve conflict before changing the clip's range
            # to preserve the current order of clips.
            self.track._resolve_clip_conflict(self.get_id(), new_clip_start_tick, new_clip_end_tick)

        # The track's clip order will be invalidated after the move, and deleting the clip
        # requires the clips to be sorted, so delete the clip before the move happens.
        if self.track is not None:
            clip_index = self.track.get_clip_index(self)
            self.track.delete_clip_at(clip_index, delete_associated_track_automation=False)

        # Move the clip.
        original_start_tick = self.get_clip_start_tick()
        original_end_tick = self.get_clip_end_tick()

        if self.get_type() == ClipType.MIDI_CLIP:
            self._proto.clip_start_tick = new_clip_start_tick
            self._proto.clip_end_tick = new_clip_end_tick
            for note in self.get_raw_notes():
                note.set_start_tick(note.get_start_tick() + offset_tick)
                note.set_end_tick(note.get_end_tick() + offset_tick)

        elif (self.get_type() == ClipType.AUDIO_CLIP):
            if (not self.has_audio_clip_data()):
                raise Exception('Cannot move audio clip without audio data')

            # Let the content of the audio at the original clip start be S
            # Let the clip start tick before move be T1, after move be T2.
            # Moving an audio clip means:
            # 1. S must be at T2 after move
            # 2. Duration of the playable audio must remain unchanged.
            song = self.song
            # The clip's end position's time relative to the audio's start time should remain
            # unchanged.
            original_audio_start_time = song.tick_to_seconds(self.get_audio_clip_data().start_tick)  # type:ignore
            original_start_time = song.tick_to_seconds(original_start_tick)
            original_end_time = song.tick_to_seconds(original_end_tick)
            playable_audio_duration = original_end_time - original_start_time
            clip_start_relative_to_audio_start = original_start_time - original_audio_start_time
            # Use the raw updated clip tick here.
            new_clip_start_time = song.tick_to_seconds(self.get_clip_start_tick() + offset_tick)
            new_audio_start_time = new_clip_start_time - clip_start_relative_to_audio_start
            new_clip_end_time = new_clip_start_time + playable_audio_duration
            self._proto.clip_start_tick = new_clip_start_tick
            self._proto.clip_end_tick = song.seconds_to_tick(new_clip_end_time)
            self._proto.audio_clip_data.start_tick = song.seconds_to_tick(new_audio_start_time)
            if (self.get_clip_end_tick() < 0):
                self.delete_from_parent(delete_associated_track_automation=True)
                return

        if (self.track is not None):
            # Ordered insert back the clip.
            self.track._ordered_insert_clip(self)

            # Move track automation points if required.
            if (move_associated_track_automation_points):
                self.track.get_automation().move_all_points_within_range(
                    original_start_tick,
                    original_end_tick,
                    offset_tick,
                    offset_value=0,
                )

    def move_clip_to(self, tick: int, move_associated_track_automation_points: bool):
        '''
        Move the clip to a given tick.

        @param tick The tick that this clip will start at.
        '''
        offsetTick = tick - self.get_clip_start_tick()
        self.move_clip(offsetTick, move_associated_track_automation_points)

    @staticmethod
    def _get_notes_in_range(raw_notes: List[song_pb2.Note], start_tick: int, end_tick: int):
        '''
        Gets notes that are playable in a range.

        Performs two O(logn) searches and some extra while loops to narrow down range.
        '''
        # Find the first note that is within the clip.
        # IMPORTANT: Assuming that a note cannot be within the Clip if it starts before the clip.
        target_start_note = SimpleNamespace()
        target_start_note.start_tick = start_tick
        start_index = max(
            0,
            lower_than(
                raw_notes,
                target_start_note,
                key=lambda x: x.start_tick,
            ),
        )
        while (
            start_index >= 0 and start_index < len(raw_notes) and
            not Clip.is_note_in_clip(
                raw_notes[start_index].start_tick,
                raw_notes[start_index].end_tick,
                start_tick,
                end_tick,
            )
        ):
            start_index += 1

        if (start_index >= len(raw_notes)):
            return []

        # Find the last note that is within the clip.
        target_end_note = SimpleNamespace()
        target_end_note.start_tick = end_tick
        end_index = min(
            len(raw_notes) - 1,
            greater_than(
                raw_notes,
                target_end_note,
                lambda x: x.start_tick
            ),
        )
        while (
            end_index >= 0 and end_index < len(raw_notes) and
            not Clip.is_note_in_clip(
                raw_notes[end_index].start_tick,
                raw_notes[end_index].end_tick,
                start_tick,
                end_tick,
            )
        ):
            end_index -= 1

        if (end_index < 0):
            return []

        if (end_index < start_index):
            return []

        return raw_notes[start_index:(end_index + 1)]

    def _trim_conflict_part(self, start_tick: int, end_tick: int):
        '''
        Trim the conflict part from the clip.

        Anything within the given range (inclusive), will be removed from the current clip.
        '''
        overlapping_start_tick = max(start_tick, self.get_clip_start_tick())
        overlapping_end_tick = min(end_tick, self.get_clip_end_tick())
        if (overlapping_start_tick > overlapping_end_tick):
            # Overlapping range is empty.
            return
        if (
            overlapping_start_tick > self.get_clip_end_tick() or
            overlapping_end_tick < self.get_clip_start_tick()
        ):
            # No overlapping.
            return

        if (
            overlapping_end_tick >= self.get_clip_end_tick() and
            overlapping_start_tick <= self.get_clip_start_tick()
        ):
            # The whole clip overlaps with the given range.
            # Delete the clip.
            self.delete_from_parent(delete_associated_track_automation=True)
            return

        if (
            overlapping_end_tick < self.get_clip_end_tick() and
            overlapping_start_tick > self.get_clip_start_tick()
        ):
            # Overlapping part is in the middle of the clip.
            # Trim the clip into the left part and create a new clip for the right part.
            right_clip_start_tick = overlapping_end_tick + 1
            right_clip_end_tick = self.get_clip_end_tick()

            self.adjust_clip_right(
                overlapping_start_tick - 1, resolve_conflict=False)

            if self.track is not None:
                if self.get_type() == ClipType.MIDI_CLIP:
                    right_clip = self.track.create_midi_clip(
                        clip_start_tick=right_clip_start_tick,
                        clip_end_tick=right_clip_end_tick,
                    )

                    right_note_protos = Clip._get_notes_in_range(
                        raw_notes=self._proto.notes, start_tick=right_clip_start_tick, end_tick=right_clip_end_tick)
                    for right_note_proto in right_note_protos:
                        right_clip.create_note(
                            pitch=right_note_proto.pitch,
                            velocity=right_note_proto.velocity,
                            start_tick=right_note_proto.start_tick,
                            end_tick=right_note_proto.end_tick,
                            update_clip_range=False,
                            resolve_clip_conflict=False,
                        )

                elif (self.get_type() == ClipType.AUDIO_CLIP):
                    audio_clip_data = self._proto.audio_clip_data
                    self.track.create_audio_clip(
                        clip_start_tick=right_clip_start_tick,
                        clip_end_tick=right_clip_end_tick,
                        audio_clip_data={
                            "audio_file_path": audio_clip_data.audio_file_path if audio_clip_data.HasField("audio_file_path") else None,
                            "start_tick": audio_clip_data.start_tick,
                            "duration": audio_clip_data.duration,
                            "audio_data": {
                                "format": audio_clip_data.audio_data.format,
                                "data": audio_clip_data.audio_data.data
                            } if audio_clip_data.HasField("audio_data") else None
                        },
                    )
            return

        # Overlapping part is on the side.
        if (
            overlapping_start_tick > self.get_clip_start_tick() and
            overlapping_start_tick <= self.get_clip_end_tick()
        ):
            # Do not resolve conflict here since it will
            # cause a bad loop.
            self.adjust_clip_right(
                overlapping_start_tick - 1, resolve_conflict=False)
        elif (
            overlapping_end_tick < self.get_clip_end_tick() and
            overlapping_end_tick >= self.get_clip_start_tick()
        ):
            # Do not resolve conflict here since it will
            # cause a bad loop.
            self.adjust_clip_left(overlapping_end_tick +
                                  1, resolve_conflict=False)

    def _get_next_note_id(self):
        if self._next_note_id is None:
            if len(self._proto.notes) == 0:
                self._next_note_id = 1
            else:
                self._next_note_id = max([note_proto.id for note_proto in self._proto.notes]) + 1

        note_id = self._next_note_id
        if (self._next_note_id >= 2147483647):
            self._next_note_id = 1
        else:
            self._next_note_id += 1

        return note_id

    def _ordered_insert_note(self, new_note: Note):
        if (self.get_type() != ClipType.MIDI_CLIP):
            # Only MIDI clips can insert notes.
            return
        if (new_note.get_clip() == self):
            # Do not insert if the note is already in the list.
            return
        insert_index = greater_equal(
            self._proto.notes,
            new_note._proto,
            key=lambda x: x.start_tick,
        )
        if (insert_index < 0):
            self._proto.notes.append(new_note._proto)
            # Reassign proto since protobuf created a new copy
            new_note._proto = self._proto.notes[-1]
        else:
            self._proto.notes.insert(insert_index, new_note._proto)
            # Reassign proto since protobuf created a new copy
            new_note._proto = self._proto.notes[insert_index]
        new_note.clip = self

    def _get_note_index(self, note: Note):
        start_index = lower_than(
            self._proto.notes,
            note._proto,
            lambda x: x.start_tick,
        )

        for index in range(max(0, start_index), len(self._proto.notes)):
            current_note_proto = self._proto.notes[index]
            if current_note_proto.id == note.get_id():
                return index
            if (current_note_proto.start_tick > note.get_start_tick()):
                # We have searched past all possible notes.
                break
        return -1

    def __repr__(self) -> str:
        return str(self._proto)

    @staticmethod
    def is_note_in_clip(
        note_start_tick: int,
        note_end_tick: int,
        clip_start_tick: int,
        clip_end_tick: int,
    ):
        '''
        Note is in the clip when:
        * The note starts at or after the clip start.
        * The note has overlapping part with the clip.
        * The overlapping part is greater than 1 tick, i.e. the note does not start from the end of the clip or end at the start of the clip.
        '''
        # If the clip starts at 0, notes that start before 0 will start at 0.
        return (
            (note_start_tick >= clip_start_tick or (clip_start_tick == 0 and note_start_tick <= 0)) and
            note_start_tick < clip_end_tick and
            note_end_tick > note_start_tick
        )

    @staticmethod
    def _generate_clip_id():
        return generate_nanoid(size=10)



from __future__ import annotations
from tuneflow_py.models.protos import song_pb2
from tuneflow_py.utils import get_audio_plugin_tuneflow_id, are_tuneflow_ids_equal, are_tuneflow_ids_equal_ignore_version, decode_audio_plugin_tuneflow_id
from nanoid import generate as generate_nanoid


class AudioPlugin:
    def __init__(self,
                 name: str | None = None,
                 manufacturer_name: str | None = None,
                 plugin_format_name: str | None = None,
                 plugin_version: str | None = None,
                 proto: song_pb2.AudioPluginInfo | None = None
                 ):
        '''
        DO NOT call the constructor directly, use Track.create_audio_plugin(tf_id: str) instead.
        '''
        if proto is not None:
            self._proto = proto
        else:
            self._proto = song_pb2.AudioPluginInfo()
            self._proto.tf_id = get_audio_plugin_tuneflow_id(
                manufacturer_name=manufacturer_name, plugin_format_name=plugin_format_name, plugin_name=name,
                plugin_version=plugin_version)
            self._proto.is_enabled = True
            self._proto.local_instance_id = AudioPlugin._generate_audio_plugin_instance_id()

    def get_tuneflow_id(self):
        '''
        A unique id that TuneFlow uses to identify the type of plugin
        '''
        return self._proto.tf_id

    def matches_tf_id(self, tf_id: str):
        '''
        Exactly matches the plugin type specified by the tf_id.
        '''
        return are_tuneflow_ids_equal(tf_id, self.get_tuneflow_id())

    def matches_tf_id_ignore_version(self, tf_id: str):
        '''
        Similar to matches_tf_id but does not check version.
        '''
        return are_tuneflow_ids_equal_ignore_version(tf_id, self.get_tuneflow_id())

    def get_instance_id(self):
        '''
        A unique id to identify the plugin instance.
        '''
        return self._proto.local_instance_id

    def to_json(self):
        audio_plugin_info = decode_audio_plugin_tuneflow_id(self.get_tuneflow_id())
        return {
            "name": audio_plugin_info["name"],
            "manufacturer_name": audio_plugin_info["manufacturer_name"],
            "plugin_format_name": audio_plugin_info["plugin_format_name"],
            "plugin_version": audio_plugin_info["plugin_version"],
            "is_enabled": self._proto.is_enabled,
        }

    def set_is_enabled(self, is_enabled: bool):
        self._proto.is_enabled = is_enabled

    def get_is_enabled(self):
        return self._proto.is_enabled

    def set_base64_states(self, base64_states: str | None = None):
        if base64_states is None:
            self._proto.ClearField("base64_states")
        else:
            self._proto.base64_states = base64_states

    def get_base64_states(self):
        return self._proto.base64_states

    def __repr__(self) -> str:
        return str(self._proto)

    @staticmethod
    def _generate_audio_plugin_instance_id():
        return generate_nanoid(size=10)

    DEFAULT_SYNTH_TFID = get_audio_plugin_tuneflow_id('TuneFlow', 'VST3', 'TFSynth', '1.0.0')



from __future__ import annotations
from tuneflow_py.descriptors.text import LabelText
from tuneflow_py.descriptors.widget import WidgetDescriptor
from tuneflow_py.descriptors.common import ClipInfo
from tuneflow_py.descriptors.clip_descriptor import AudioData
from typing import Any, List
from typing_extensions import TypedDict, Required, NotRequired, Literal
from enum import Enum


class Mp3DataConvertOptions(TypedDict):
    sampleRate: NotRequired[int]
    ''' Sample rate in Hz, default 44100. '''


class AudioDataConvertOptions(TypedDict):
    toFormat: Required[Literal['ogg', 'wav']]
    options: NotRequired[Mp3DataConvertOptions]


class AudioDataInjectOptions(TypedDict):
    convert: NotRequired[AudioDataConvertOptions]
    '''
    When enabled, only the visible portion of the clip will be converted to the given format.
   
    The converted audio data will start from the clip start and end at the clip end.
    '''


class ClipAudioDataInjectOptions(AudioDataInjectOptions):
    '''
    Inject config for when injection source is `InjectSource.ClipAudioData`.
    '''
    clips: Required[Literal['selectedAudioClips',] | List[ClipInfo]]


class SelectedClipInfosInjectOptions(TypedDict):
    '''
    Inject config for when injection source is `InjectSource.SelectedClipInfos`.
    '''

    maxNumClips: NotRequired[int]
    '''
    Maximum number of clip infos to include. Defaults to unlimited.
    '''


class InjectConfig(TypedDict):
    '''
    Used to specify a injection source when additional config is needed.
    '''

    type: Required[int]
    options: Required[ClipAudioDataInjectOptions | SelectedClipInfosInjectOptions]


class InjectSource(Enum):
    SelectedTrackIds = 1
    ''' A string that represents the ids of the current selected tracks. '''

    SelectedClipInfos = 2
    ''' A `list[ClipInfo]` that represents the `ClipInfo`s of the current selected clips. '''

    TickAtPlayhead = 3
    ''' A number that represents the current tick that the playhead is at. '''

    EditingClipInfo = 4
    ''' The `ClipInfo` of the clip that is being edited in the MIDI editor. '''

    EditingNoteIds = 5
    ''' The ids of the notes that is being edited in the MIDI editor. '''

    TickAtPlayheadSnappedToBeat = 6
    ''' A number that represents the start tick of the beat where the playhead is at. '''

    ClipAudioData = 7
    ''' A list of `AudioData` for specified clips. '''


class ParamDescriptor(TypedDict):
    '''
    Definition of the config for a single param.

    Note that the members here use camel case to be
    consistent with the typescript plugins.
    '''

    displayName: Required[LabelText]
    ''' The name to display on the UI. '''

    widget: Required[WidgetDescriptor]
    '''
    Configuration of the widget to display on the UI.

    Use  `{ "type": WidgetType.NoWidget.value }` if you don't need a widget.
    '''

    defaultValue: NotRequired[Any]
    ''' The default value of the param. '''

    adjustable: NotRequired[bool]
    '''
    Whether this param is adjustable. If the param is not adjustable, the controllable
    part is not shown.

    Defaults to true.
    '''

    hidden: NotRequired[bool]
    ''' Whether this param is hidden completely. Defaults to false. '''

    optional: NotRequired[bool]
    ''' Whether this param can be left None. Defaults to false. '''

    description: NotRequired[LabelText]
    ''' Explaining what this parameter is for. '''

    injectFrom: NotRequired[InjectSource | InjectConfig]
    '''
    Injects the value from the editor at the time the plugin runs.
    
    If specified, the editor will inject the value specified by the `InjectSource` or `InjectConfig`,
    '''

    adjustableWhenPluginIsApplied: NotRequired[bool]
    '''
    If set to true, this param can still be adjusted
    when the plugin is applied. However, changing the param
    will invalidate all the changes after this plugin.

    Defaults to false.
    '''


class ClipAudioDataInjectDataEntry(TypedDict):
    clipInfo: Required[ClipInfo]
    audioData: Required[AudioData]


ClipAudioDataInjectData = List[ClipAudioDataInjectDataEntry]
'''
Type of the injected data when injection source is `InjectSource.ClipAudioData`.
'''
